//
// BasicString.cc
//
//	Contains library of functions for handling strings
//	for the VAX-BASIC translator.
//
// Author:
//	04/02/1996 - Kevin Handy
//

//
//	Codes for data type, dsc$b_dtype:
//
//	Atomic data types:

static const unsigned char DSC$K_DTYPE_Z = 0;	// unspecified
static const unsigned char DSC$K_DTYPE_BU = 2;	// byte logical;  8-bit unsigned quantity
static const unsigned char DSC$K_DTYPE_WU = 3;	// word logical;  16-bit unsigned quantity
static const unsigned char DSC$K_DTYPE_LU = 4;	// longword logical;  32-bit unsigned quantity
static const unsigned char DSC$K_DTYPE_QU = 5;	// quadword logical;  64-bit unsigned quantity
static const unsigned char DSC$K_DTYPE_OU = 25;	// octaword logical;  128-bit unsigned quantity
static const unsigned char DSC$K_DTYPE_B = 6;	// byte integer;  8-bit signed 2's-complement integer
static const unsigned char DSC$K_DTYPE_W = 7;	// word integer;  16-bit signed 2's-complement integer
static const unsigned char DSC$K_DTYPE_L = 8;	// longword integer;  32-bit signed 2's-complement integer
static const unsigned char DSC$K_DTYPE_Q = 9;	// quadword integer;  64-bit signed 2's-complement integer
static const unsigned char DSC$K_DTYPE_O = 26;	// octaword integer;  128-bit signed 2's-complement integer
static const unsigned char DSC$K_DTYPE_F = 10;	// F_floating;	32-bit single-precision floating point
static const unsigned char DSC$K_DTYPE_D = 11;	// D_floating;	64-bit double-precision floating point
static const unsigned char DSC$K_DTYPE_G = 27;	// G_floating;	64-bit double-precision floating point
static const unsigned char DSC$K_DTYPE_H = 28;	// H_floating;	128-bit quadruple-precision floating point
static const unsigned char DSC$K_DTYPE_FC = 12;	// F_floating complex
static const unsigned char DSC$K_DTYPE_DC = 13;	// D_floating complex
static const unsigned char DSC$K_DTYPE_GC = 29;	// G_floating complex
static const unsigned char DSC$K_DTYPE_HC = 30;	// H_floating complex
static const unsigned char DSC$K_DTYPE_CIT = 31; // COBOL Intermediate Temporary

//	String data types:
static const unsigned char DSC$K_DTYPE_T = 14;	// character-coded text;  a single character or a string
static const unsigned char DSC$K_DTYPE_T2 = 38;	// 16-bit Text (Asian workstations)
static const unsigned char DSC$K_DTYPE_VT = 37;	// varying character-coded text;  16-bit count, followed by a string
static const unsigned char DSC$K_DTYPE_NU = 15;	// numeric string, unsigned
static const unsigned char DSC$K_DTYPE_NL = 16;	// numeric string, left separate sign
static const unsigned char DSC$K_DTYPE_NLO = 17; // numeric string, left overpunched sign
static const unsigned char DSC$K_DTYPE_NR = 18;	// numeric string, right separate sign
static const unsigned char DSC$K_DTYPE_NRO = 19; // numeric string, right overpunched sign
static const unsigned char DSC$K_DTYPE_NZ = 20;	// numeric string, zoned sign
static const unsigned char DSC$K_DTYPE_P = 21;	// packed decimal string
static const unsigned char DSC$K_DTYPE_V = 1;	// bit;  aligned bit string
static const unsigned char DSC$K_DTYPE_VU = 34;	// bit unaligned;  arbitrary bit string

//	Miscellaneous data types:
static const unsigned char DSC$K_DTYPE_ZI = 22;	// sequence of instructions
static const unsigned char DSC$K_DTYPE_ZEM = 23; // procedure entry mask
static const unsigned char DSC$K_DTYPE_DSC = 24; // descriptor
static const unsigned char DSC$K_DTYPE_BPV = 32; // bound procedure value
static const unsigned char DSC$K_DTYPE_BLV = 33; // bound label value
static const unsigned char DSC$K_DTYPE_ADT = 35; // absolute date and time

//	Introduced with Alpha-AXP:
static const unsigned char DSC$K_DTYPE_FS = 52;	// 32-bit S_float IEEE single precision
static const unsigned char DSC$K_DTYPE_FT = 53;	// 64-bit T_float IEEE double precision
static const unsigned char DSC$K_DTYPE_FSC = 54; // 64-bit S_complex IEEE <S-real,S-imag>
static const unsigned char DSC$K_DTYPE_FTC = 55; // 128-bit T_complex IEEE
static const unsigned char DSC$K_DTYPE_FX = 57;	// 128-bit X_float IEEE quadruple precision
static const unsigned char DSC$K_DTYPE_FXC = 58; // 256-bit X_complex IEEE

//	Reserved data type codes:
//		 39-177 are reserved to DEC;
//		192-255 are reserved to CSS and customers.


//
//	Descriptor Class Codes for dsc$b_class:
//
static const unsigned char DSC$K_CLASS_S = 1;	// scalar or string descriptor
static const unsigned char DSC$K_CLASS_D = 2;	// dynamic string descriptor
static const unsigned char DSC$K_CLASS_V = 3;	// variable buffer descriptor;	reserved for use by DIGITAL
static const unsigned char DSC$K_CLASS_A = 4;	// array descriptor
static const unsigned char DSC$K_CLASS_P = 5;	// procedure descriptor
static const unsigned char DSC$K_CLASS_PI = 6;	// procedure incarnation descriptor;  obsolete
static const unsigned char DSC$K_CLASS_J = 7;	// label descriptor;  reserved for use by the VAX-11 Debugger
static const unsigned char DSC$K_CLASS_JI = 8;	// label incarnation descriptor;  obsolete
static const unsigned char DSC$K_CLASS_SD = 9;	// decimal scalar string descriptor
static const unsigned char DSC$K_CLASS_NCA = 10; // noncontiguous array descriptor
static const unsigned char DSC$K_CLASS_VS = 11;	// varying string descriptor
static const unsigned char DSC$K_CLASS_VSA = 12; // varying string array descriptor
static const unsigned char DSC$K_CLASS_UBS = 13; // unaligned bit string descriptor
static const unsigned char DSC$K_CLASS_UBA = 14; // unaligned bit array descriptor
static const unsigned char DSC$K_CLASS_SB = 15;	// string with bounds descriptor
static const unsigned char DSC$K_CLASS_UBSB = 16; // unaligned bit string with bounds descriptor

//	Reserved descriptor class codes:
//		 17-190 are reserved to DEC;
//		192-255 are reserved to CSS and customers.



//
// Base class for all types of strings.
// Should be compatible with descriptors under VMS, but don't
// use any VMS library calls since we want to be able to run under
// Un*x clones.
//
class BasicDescriptor
{
private:
	unsigned short  dsc$w_length;	// specific to descriptor class;
					//  typically a 16-bit (unsigned) length
	unsigned char   dsc$b_dtype;    // data type code
	unsigned char   dsc$b_class;    // descriptor class code
	void            *dsc$a_pointer; // address of first byte of data element

public:
	BasicDescriptor();
	BasicDescriptor(char* string);
	~BasicDescriptor();

};


class BasicString : public BasicDescriptor
{
public:
	BasicString();
	~BasicString();
}


//
// Implementation of BasicDescriptor
//
BasicDescriptor::BasicDescriptor()
{
	dsc$w_length = 0;
	dsc$b_dtype = 0;
	dsc$b_class = 0;
	dsc$a_pointer = NULL;
}

//
// Implementation of BasicString
//
BasicString::BasicString(char* string)
{
	dsc$w_length = strlen(string);
	dsc$b_dtype = DSC$K_DTYPE_T;
	dsc$b_class = DSC$K_CLASS_S;
	dsc$a_pointer = string;
}

