
long ErrNum = 0;
long ErrLine = 0;
long Status = 0;

//
// bstring.h
//
//	String class for Basic to C++ translator
//
// Description:
//
//	This file impliments a string class for my Basic
//	translater using VMS style descriptors and string
//	functions. This allows the programs to interface
//	with the VMS routines as easily as under the
//	real VAX/BASIC compiler.
//
//	This is really nothing more than a wrapper around
//	the STR$xxx functions, and the DSC$DESCRIPTOR_S
//	stuff.
//
// History:
//
//	01.00.00 - 09/19/1996 - Kevin Handy
//		Original version. Very minimal. Assignment and output
//		only exist. All code contained in '.h' file (easier
//		to debug).
//
//	01.02.01 - 09/20/1996 - Kevin Handy
//		Added in '+' operator for BString and char.
//
//	01.02.02 - 01/01/1997 - Kevin Handy
//		Added sever string functions, including left,
//		right, pos, mid, string, space, strlen.
//
//	01.02.03 - 01/04/1997 - Kevin Handy
//		Added comparison operators (Bstring to Bstring)
//
//	01.02.04 - 01/05/1997 - Kevin Handy
//		Added more functions. Qnum, Qnum1
//
//	01.02.05 - 01/26/1997 - Kevin Handy
//		Code for gosub.
//
//	01.03.01 - 01/29/1997 - Kevin Handy
//		More functions! Ascii, Edit, Ert.
//
//	01.03.02 - 02/07/1997 - Kevin Handy
//		Change 'short' to 'unsigned short' for various length
//		variable definitions.
//
//	01.03.03 - 02/19/1997 - Kevin Handy
//		Use str$compare instead of str$compare_eql in
//		operator '==', so matches work like Vax/Basic
//		(re spaces at end of a string)
//
//	01.03.04 - 02/19/1997 - Kevin Handy
//		Fix a couple of bugs in 'edit' function.
//
//	01.03.05 - 04/09/1997 - Kevin Handy
//		Move out stuff not related to string into basicfun.h
//		Modify "char" based "instr" function to handle base
//		too small problems.

//
// Include files
//
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include "bstring.h"
#include "descrip.h"
#include "str$routines.h"
#include <iostream.h>

//
// Empty Constructor
//
BString::BString(void)
{
	dsc$b_dtype = DSC$K_DTYPE_T;		// Text class
	dsc$b_class = DSC$K_CLASS_D;		// Dynamic
	dsc$w_length = 0;			// Length = 0
	dsc$a_pointer = NULL;			// No allocation
}

//
// Create from a text string
//
BString::BString(const char* Text)
{
	unsigned short Length = strlen(Text);

	dsc$b_dtype = DSC$K_DTYPE_T;		// Text class
	dsc$b_class = DSC$K_CLASS_D;		// Static
	dsc$w_length = 0;			// Length
	dsc$a_pointer = NULL;			// Point at origional string

	str$copy_r(this, &Length, Text);
}

//
// Create from a text string
//
BString::BString(const char* Text, unsigned short Length)
{
	dsc$b_dtype = DSC$K_DTYPE_T;		// Text class
	dsc$b_class = DSC$K_CLASS_D;		// Static
	dsc$w_length = 0;			// Length
	dsc$a_pointer = NULL;			// Point at origional string

	str$copy_r(this, &Length, Text);
}

//
// Create from a byte constant
//
BString::BString(char ch)
{
	const unsigned short Length = 1;
	char Text[2];

	dsc$b_dtype = DSC$K_DTYPE_T;		// Text class
	dsc$b_class = DSC$K_CLASS_D;		// Static
	dsc$w_length = 0;			// Length
	dsc$a_pointer = NULL;			// Pointer

	Text[0] = ch;
	Text[1] = '\0';
	str$copy_r(this, &Length, &Text);
}


//
// Destructor
//
BString::~BString()
{
// cout << "Killed" << endl;
	//
	// Release memory allocation if this is a string class
	//
	if ((dsc$b_class == DSC$K_CLASS_D) && (dsc$a_pointer != NULL))
	{
		str$free1_dx(this);
	}
}

//
// Assignment
//
BString& BString::operator = (const char* Text)
{
	unsigned short Length = strlen(Text);

	str$copy_r(this, &Length, Text);
	return *this;
}

//
// Left()
//
BString left(const BString& Source, long Length)
{
	BString* NewString = new BString;
	str$left(NewString, &Source, &Length);
	return *NewString;
}

//
// Right()
//
BString right(const BString& Source, long Length)
{
	BString* NewString = new BString;
	str$right(NewString, &Source, &Length);
	return *NewString;
}

//
// Pos()
//
BString Qseg(const BString& Source, long Start, long End)
{
	BString* NewString = new BString;
	str$pos_extr(NewString, &Source, &Start, &End);
	return *NewString;
}

//
// Mid()
//
BString mid(const BString& Source, long Start, long Length)
{
	BString* NewString = new BString;
	str$len_extr(NewString, &Source, &Start, &Length);
	return *NewString;
}

//
// string
//
BString Qstring(long Count, int Character)
{
	BString* NewString = new BString;
	char Source = Character;

	str$dupl_char(NewString, &Count, &Source);

	return *NewString;
}

//
// char
//
BString Char(int Character)
{
	char Temp[2];

	Temp[0] = Character;
	Temp[1] = '\0';

	BString* NewString = new BString(Temp);

	return *NewString;
}

//
// space
//
BString Qspace(long Count)
{
	BString* NewString = new BString;
	char XSpace = ' ';

	str$dupl_char(NewString, &Count, &XSpace);

	return *NewString;
}

//
// Qnum(integer)
//
BString Qnum(long Number)
{
	char Result[32];
	BString* NewString = new BString;

	sprintf(Result, "%ld ", Number);
	*NewString = Result;

	return *NewString;
}

//
// Qnum(float)
//
BString Qnum(double Number)
{
	char Result[32];
	BString* NewString = new BString;

	sprintf(Result, "%f ", Number);
	*NewString = Result;

	return *NewString;
}

//
// Qnum1(integer)
//
BString Qnum1(long Number)
{
	char Result[32];
	BString* NewString = new BString;

	sprintf(Result, "%ld", Number);
	*NewString = Result;

	return *NewString;
}

//
// Qnum1(float)
//
BString Qnum1(double Number)
{
	char Result[64];
	BString* NewString = new BString;

	sprintf(Result, "%f", Number);
	*NewString = Result;

	return *NewString;

}

//
// Trm
//
BString trm(const BString& Source)
{
	BString* NewString = new BString;
	unsigned short Length;

	str$trim(NewString, &Source, &Length);
	return *NewString;
}


//
// strlen
//
long strlen(const BString& Source)
{
	unsigned short Length;
	char* Address;

	str$analyze_sdesc(&Source, &Length, &Address);
	return Length;
}


//
// Addition
//
BString& operator + (const BString& Text1, const BString& Text2)
{
	//
	// Allocate a new string
	//
	BString* NewString = new BString;

	//
	// Combine strings together.
	//	Using copy,aappend instead of concat, because varargs
	//	won't give me information about number of paramaters.
	//
	str$copy_dx(NewString, &Text1);
	str$append(NewString, &Text2);

	//
	// Give back the new string
	//
	return *NewString;
}

BString& operator + (const BString& Text1, const char* Text2)
{
	//
	// Allocate a new string
	//
	BString* NewString = new BString;
	BString Transform(Text2);

	//
	// Combine strings together
	//
	str$copy_dx(NewString, &Text1);
	str$append(NewString, &Transform);

	//
	// Give back the string
	//
	return *NewString;
}

//
// Various string type functions
//
int Ascii(const char x)
{
	return x;
}

int Ascii(const char* x)
{
	return *x;
}

int Ascii(const BString& Source)
{
	unsigned short Length;
	char* Address;
	str$analyze_sdesc(&Source, &Length, &Address);
	if (Length == 0)
	{
		return 0;
	}
	else
	{
		return *Address;
	}
}

//
// rad()
//
BString& Qrad(const unsigned int Number)
{
	const char RadChar[] = " ABCDEFGHIJKLMNOPQRSTUVWXYZ$.?0123456789";
	char Result[4];
	BString* NewString = new BString;

	Result[0] = RadChar[(Number / (40 * 40)) % 40];	// 1st Character
	Result[1] = RadChar[(Number / 40) % 40];	// 2nd Character
	Result[2] = RadChar[Number % 40];		// 3rd character
	Result[3] = '\0';

	*NewString = Result;
	return *NewString;
}

//
// Ert
//	This is a really lousy stub for ert$
//
BString& ert(const long Number)
{
	BString* NewString = new BString("Error Message #000000");
	return *NewString;
}

//
// Instr
//
int instr(int Start, char* Base, char* Search)
{
	if (Start > strlen(Base))
	{
		return 0;
	}
	else
	{
		char* test = strstr((Base + Start - 1), Search);
		return (test != NULL) ? (int)(test - Base + 1) : 0;
	}
}

int instr(long Start, const BString& BaseString,
	const BString& SearchString)
{
	return str$position(&BaseString, &SearchString, &Start);
}

//
// Edit
//	Also used by cvt$$ code.
//
BString& edit(const BString& Source, const int Code)
{
	BString* NewString = new BString;
	unsigned short Length;
	char* Address;
	int ResultPtr = 0;
	int LeadSpace = 0;
	int MultiSpace = 0;
	int QuoteMark = 0;

	str$analyze_sdesc(&Source, &Length, &Address);
	char* Result = new char[Length + 1];

	while (Length--)
	{
		unsigned char OneCh = *Address++;

		//
		// 1 - Discard parity bit
		//
		if (Code & 1)
		{
			OneCh &= 0x7f;
		}

		//
		// 2 - Discard all spaces and tabs
		//
		if ((Code & 2) && ((OneCh == ' ') || (OneCh == '\t')))
		{
			if (QuoteMark == 0)
			{
				continue;
			}
		}

		//
		// 4 - Discard all carriage returns, line feeds,
		// form feeds, deletes, escapes, and nulls
		//
		if ((Code & 4) && ((OneCh == '\r') || (OneCh == '\n') ||
			(OneCh == '\f') || (OneCh == 27) || (OneCh == 0)))
		{
			if (QuoteMark == 0)
			{
				continue;
			}
		}

		//
		// 8 - Discard Leading Spaces and Tabs
		//
		if ((Code & 8) && (LeadSpace == 0))
		{
			if ((OneCh == ' ') || (OneCh == '\t'))
			{
				continue;
			}
			LeadSpace = 1;
		}

		//
		// 16 - Convert Multiple Spaces and Tabs to one space
		//
		if (Code & 16)
		{
			if ((OneCh == ' ') || (OneCh == '\t'))
			{
				OneCh = ' ';
				if (MultiSpace != 0)
				{
					continue;
				}
				MultiSpace = 1;
			}
			else
			{
				MultiSpace = 0;
			}
		}

		//
		// 32 - Convert lower case to upper case
		//
		if (Code & 32)
		{
			if (QuoteMark == 0)
			{
				OneCh = toupper(OneCh);
			}
		}

		//
		// 64 - Convert left bracket to left parentheses, and
		// right bracket to right parentheses
		//
		if ((Code & 64) && (QuoteMark == 0))
		{
			if (OneCh == '[')
			{
				OneCh = '(';
			}
			if (OneCh == ']')
			{
				OneCh = ')';
			}
		}

		//
		// 256 - Suppress all editing for characters within
		// quotation marks. If the string has only one quotation
		// mark, suppresses all editing for the characters
		// following the quotation mark.
		//
		if (Code & 256)
		{
			if (QuoteMark)
			{
				if (OneCh == QuoteMark)
				{
					QuoteMark = 0;
				}
			}
			else
			{
				if ((OneCh == '"') || (OneCh == '\''))
				{
					QuoteMark = OneCh;
				}
			}
		}

		//
		// If we get here, we want to keep the character
		//
		Result[ResultPtr++] = OneCh;
	}

	//
	// 128 - Discard Trailing Spaces and Tabs
	//
	if ((Code & 128) && (QuoteMark == 0))
	{
		while ((ResultPtr > 0) &&
			((Result[ResultPtr - 1] == ' ') ||
			(Result[ResultPtr - 1] == '\t')))
		{
			ResultPtr--;
		}
	}

	//
	// Terminate the string
	//
	Result[ResultPtr] = '\0';

	//
	// Return the final answer
	//
	*NewString = Result;
	return *NewString;
}

double atof(const BString& Source)
{
	unsigned short Length;
	char* Address;
	double result;

	//
	// Determine string info
	//
	str$analyze_sdesc(&Source, &Length, &Address);

	//
	// make a copy we can do stuff to
	//
	if (Length == 0)
	{
		return 0.0;
	}
	char* Work = new char[Length + 1];
	assert(Work != NULL);

	memcpy(Work, Address, Length);
	Work[Length] = '\n';

	//
	// Now we can actually do an atof
	//
	result = atof(Work);

	//
	// Release allocated memory
	//
	delete Work;

	//
	// Return result
	//
	return result;
}

