//
// bstring.h
//
//	String class for Basic to C++ translator
//
// Description:
//
//	This file impliments a string class for my Basic
//	translater using VMS style descriptors and string
//	functions. This allows the programs to interface
//	with the VMS routines as easily as under the
//	real VAX/BASIC compiler.
//
//	This is really nothing more than a wrapper around
//	the STR$xxx functions, and the DSC$DESCRIPTOR_S
//	stuff.
//
// History:
//
//	01.00.00 - 09/19/1996 - Kevin Handy
//		Original version. Very minimal. Assignment and output
//		only exist. All code contained in '.h' file (easier
//		to debug).
//
//	01.02.01 - 09/20/1996 - Kevin Handy
//		Added in '+' operator for BString and char.
//
//	01.02.02 - 01/01/1997 - Kevin Handy
//		Added sever string functions, including left,
//		right, pos, mid, string, space, strlen.
//
//	01.02.03 - 01/04/1997 - Kevin Handy
//		Added comparison operators (Bstring to Bstring)
//
//	01.02.04 - 01/05/1997 - Kevin Handy
//		Added more functions. Qnum, Qnum1
//
//	01.02.05 - 01/26/1997 - Kevin Handy
//		Code for gosub.
//
//	01.03.01 - 01/29/1997 - Kevin Handy
//		More functions! Ascii, Edit, Ert.
//
//	01.03.02 - 02/07/1997 - Kevin Handy
//		Change 'short' to 'unsigned short' for various length
//		variable definitions.
//
//	01.03.03 - 02/19/1997 - Kevin Handy
//		Use str$compare instead of str$compare_eql in
//		operator '==', so matches work like Vax/Basic
//		(re spaces at end of a string)
//
//	01.03.04 - 02/19/1997 - Kevin Handy
//		Fix a couple of bugs in 'edit' function.
//
//	01.03.05 - 04/09/1997 - Kevin Handy
//		Move out stuff not related to string into basicfun.h
//		Modify "char" based "instr" function to handle base
//		too small problems.

#ifndef _bstring_h_
#define _bstring_h_

//
// Include files
//
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include "descrip.h"
#include "str$routines.h"
#include <iostream.h>

//
// BString class
//
class BString : public dsc$descriptor_s
{
public:
	BString(void);
	BString(const char* Text);
	BString(const char* Text, unsigned short Length);
	BString(char ch);
	~BString();

	//
	// Operators
	//
	BString& operator = (const char* Text);
	inline BString& operator = (const BString& Source)
		{ str$copy_dx(this, &Source); return *this; }

	//
	// Functions
	//
	void ReMap(const char* Value, int Length);

	//
	// I/O functions (Hack, needs lots of work)
	//
	inline friend ostream& operator<< (ostream& os, const BString& bs)
//		{ if (bs.dsc$a_pointer != NULL) {for (int loop=0; loop<bs.dsc$w_length; loop++)
//			os << bs.dsc$a_pointer[loop];} return os; }
		{ if (bs.dsc$a_pointer != NULL)
			{os.write(bs.dsc$a_pointer, bs.dsc$w_length); } }

	inline friend istream& operator>> (istream& is, BString& bs)
		{ char inputstr[128]; is >> inputstr; bs = inputstr; }
};


BString left(const BString& Source, long Length);
BString right(const BString& Source, long Length);
BString Qseg(const BString& Source, long Start, long End);
BString mid(const BString& Source, long Start, long Length);
BString Qstring(long Count, int Character);
BString Char(int Character);
BString Qspace(long Count);
BString Qnum(long Number);
//BString Qnum(int Number) { return Qnum((long)Number); }
BString Qnum(double Number);
BString Qnum1(long Number);
//BString Qnum1(int Number) { return Qnum1((long)Number); }
BString Qnum1(double Number);

//
// str
//
inline BString str(long Number) { return Qnum1(Number); }
inline BString str(double Number) { return Qnum1(Number); }

BString trm(const BString& Source);
long strlen(const BString& Source);
BString& operator + (const BString& Text1, const BString& Text2);
BString& operator + (const BString& Text1, const char* Text2);
inline int operator == (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) == 0;
}

inline int operator < (const BString& Text1, const BString& Text2)
{
	 return str$compare(&Text1, &Text2) == -1;
}

inline int operator <= (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) <= 0;
}

inline int operator > (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) == 1;
}

inline int operator >= (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) >= 0;
}

inline int operator != (const BString& Text1, const BString& Text2)
{
	return str$compare(&Text1, &Text2) != 0;
}

//
// Various string type functions
//
int Ascii(const char x);
int Ascii(const char* x);
int Ascii(const BString& Source);
BString& Qrad(const unsigned int Number);
BString& ert(const long Number);
int instr(int Start, char* Base, char* Search);
int instr(long Start, const BString& BaseString,
	const BString& SearchString);
BString& edit(const BString& Source, const int Code);
double atof(const BString& Source);
BString sys(const BString& Source);
BString Qdate(int x);
BString Qtime(int x);
long cvtai(BString x);
BString cvtia(int x);
double cvtaf(BString x);
BString cvtfa(double x);

#endif

