/*******************************************************************************

Program:

	nodes.cc
	01.04

Title:

	Routines to work with nodes.

Description:

	This file contains various routines for manipulating the tree
	nodes used when parsing a basic program.

Compile:

	cc [noopt]

History:

	01.01.00 - 08/30/1996 - Kevin Handy
		Update version number to 1.1, and install on internet.

	01.01.01 - 08/31/1996 - Kevin Handy
		Changed 'TreeStruct' to 'Node', which makes mode sense.

	01.01.02 - 09/19/1996 - Kevin Handy
		Always include 'bstring.h' for string variables.

	01.01.03 - 09/20/1996 - Kevin Handy
		Include <errno.h> instead of <error.h> (oops).
		Always include <stdlib.h>

	01.01.04 - 10/06/1996 - Kevin Handy
		Added BAS_V_PREDEF capability.

	01.01.05 - 10/20/1996 - Kevin Handy
		Put a semicolon ';' after labels. This eliminates the
		error "label must be followed by a statement" from G++.
		Output 'include "datalist.h"' when needed.

	01.02.01 - 11/11/1996 - Kevin Handy
		Added parameter to output functions (ostream& os),
		so that I didn't have to sneak in the channel through
		a side path. (Also loose weird *OutFile<< syntax)

	01.02.02 - 11/11/1996 - Kevin Handy
		Midifications to use "os <<" instead of calls to
		"Outputxxx(os)" whenever possible.

	01.02.03 - 11/12/1996 - Kevin Handy
		Modifications to reduce the number of functions,
		along with changes to parse.y

	01.02.04 - 12/27/1996 - Kevin Handy
		Make the INPUT statement do something again.

	01.02.05 - 01/05/1997 - Kevin Handy
		Force string constants to be BString's when necessary.

	01.03.01 - 01/26/1997 - Kevin Handy
		Work up a gosub/return code (horrible stuff).

	01.03.02 - 01/29/1997 - Kevin Handy
		Make LINPUT work exactly like INPUT for now.

	01.03.03 - 01/29/1997 - Kevin Handy
		Output '\x' instead of '\0x'.

	01.03.04 - 02/03/1997 - Kevin Handy
		Small fix to gosub code. (ONGOSUB, Clearing Flag)

	01.03.05 - 02/06/1997 - Kevin Handy
		Handle DECLARE, definitions better.
		"EXTERNAL FUNCTION" without parameters should output
		"(...)" instead of "()".
		Output ErrNum and ErrLine instead of errno and errline.

	01.03.06 - 02/09/1997 - Kevin Handy
		Work on 'MAP' and 'COMMON' some more.
		Work on 'INPUT/OUTPUT' statements.
		Fixed output of numeric constants so they don't copy over
		leadsing zeros.

	01.03.07 - 02/18/1997 - Kevin Handy
		Change LSET and RSET code from using external functions
		to using string class functions.

	01.03.08 - 02/18/1997 - Kevin Handy
		Fix LINPUT bug.

	01.03.09 - 02/18/1997 - Kevin Handy
		Start coding for FIELD statements.

	01.03.10 - 02/19/1997 - Kevin Handy
		Improve code generated for INPUT_LINE and FIELD.

	01.03.11 - 03/14/1997 - Kevin Handy
		Handle adding two string constants together to
		use C's string concatination feature.

	01.03.12 - 04/02/1997 - Kevin Handy
		Added CLUSTERSIZE option to open.

	01.03.13 - 04/06/1997 - Kevin Handy
		Modifications necessary to move DEF functoins out of
		main functions.

	01.03.14 - 04/06/1997 - Kevin Handy
		Optimize UNLESS code generated.

	01.03.15 - 04/06//1997 - Kevin Handy
		Handle BAS_X_LNOT, BAS_X_LAND, and BAS_X_LOR codes.

	01.03.16 - 04/06/1997 - Kevin Handy
		More work on data statements. Handle wrapping
		long lines, and new variant of the dataListClass
		that should work better.

	01.03.17 - 04/11/1997 - Kevin Handy
		Start on "ON ERROR" code.

	01.03.18 - 04/21/1997 - Kevin Handy
		Added code for several more BASIC constants.

	01.03.19 - 04/23/1997 - Kevin Handy
		Work on "PRINT USING" statements.

	01.03.20 - 04/24/1997 - Kevin Handy
		Modify FOR to look for UMINUS in step.

	01.03.21 - 04/28/1997 - Kevin Handy
		Put each print using format item on it's own line,
		because G++ will otherwise run through them backwards.

	01.03.22 - 04/28/1997 - Kevin Handy
		Handle special assignment (DEF return value) as a
		'return xxx' instead of a '(Return) = '.

	01.03.23 - 06/04/1997 - Kevin Handy
		Modified to handle merged data statements.

	01.03.24 - 05/09/1997 - Kevin Handy
		Added StripLeadZero function.

	01.03.25 - 06/16/1997 - Kevin Handy
		Modifications to make array definitions 1 element
		larger, because in basic 'xxx[10]' create an array
		0..10, and C++ creates 0..9

	01.03.26 - 06/16/1997 - Kevin Handy
		Modifications to handle 'main' definition better.

	01.04.00 - 06/16/1997 - Kevin Handy
		Set version to 1.4

	01.04.01 - 06/18/1997 - Kevin Handy
		In dimension statements, actually increment of the number
		instead of adding a string ' + 1' when possible.
		Added function 'IsSimpleNumber'.
		Also fixed indent bug for closing '}' of main function.

	01.04.02 - 06/19/1997 - Kevin Handy
		Change 'cout' to 'os' in a couple of places

	01.04.03 - 06/20/1997 - Kevin Handy
		Fix indent in def of Result when not outputing it in main

	01.04.04 - 06/24/1997 - Kevin Handy
		Fix 'DEF' functions to get a type other than 'void'.

	01.04.05 - 06/24/997 - Kevin Handy
		Split 'OutputDefinition' into 'OutputDefinition',
		'OutputDefinitionList' 'OutputDefinitionParam', and
		'OutputDefinitionCopy' so could get better handling.

	01.04.06 - 06/24/1997 - Kevin Handy
		Make 'Result' follow data type of function instead of
		always being defined as an int.

	01.04.07 - 07/02/1997 - Kevin Handy
		Only put '...' in function definition when defining
		an external function, use 'void' in defining an
		actual function.

	01.04.08 - 07/02/1997 - Kevin Handy
		Improve code generated for get/put/find/reset.

	01.04.09 - 07/05/1997 - Kevin Handy
		Some improvement to I/O Channels. Start of getting
		channels to work.

	01.04.10 - 07/05/1997 - Kevin Handy
		Fix RESUME so that it flags line numbers as being
		necessary.

	01.04.11 - 07/05/1997 - Kevin Handy
		Include 'unistd.h' when necessary.

	01.04.12 - 07/06/1997 - Kevin Handy
		Work on determining the VARTYPE of an expression,
		so that I can handle forcing conversions when
		necessary.

	01.04.13 - 07/06/1997 - Kevin Handy
		Force array references to be integer types.

	01.04.14 - 07/06/1997 - Kevin Handy
		Fixed DIM statesments so it doesn't define everything
		as a 'Dfloat'.

	01.04.15 - 07/07/1997 - Kevin Handy
		Handle BAS_V_INT in IsSimpleInteger function.

	01.04.16 - 07/07/1997 - Kevin Handy
		Fix 'ON xxx GOTO yyy' for syntax errors.

	01.04.17 - 07/07/1997 - Kevin Handy
		Handle 'Result' in ASSIGNLIST's like ASSIGN's.

	01.04.18 - 07/07/1997 - Kevin Handy
		Output ';' after 'unlink()'.

	01.04.19 - 07/07/1997 - Kevin Handy
		Fix several bugs (causing crash) in 'GET' statements.

	01.04.20 - 07/09/1997 - Kevin Handy
		Fixed bugs with data statements.

	01.04.21 - 07/25/1997 - Kevin Handy
		Added code to handle 'KeepAllLines' variable

	01.04.22 - 08/04/1997 - Kevin Handy
		Output function prototypes for declared functions

	01.04.23 - 08/05/1997 - Kevin Handy
		Force various function types to be globally defined,
		instead of locally. This will remove the 'duplicate'
		definitions iin other functions that call it.

	01.04.24 - 08/20/1997 - Kevin Handy
		Fix indenting problem in 'GET' code.
		Fix 'FUNCTION REAL...' so it doesn't output 'UNDEFUNED VOID'
		as the variable type.

*******************************************************************************/

//
// System Include Files
//
#include <iostream.h>
#include <stdio.h>
#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

//
// Project Include Files
//
#include "basic.h"
#include "parse.h"
#include "nodes.h"

//
// Module #defined Constants
//
#define MAXLEVEL 256                    // Maximum depth of nodes

//
// Module #defined Functions
//

//
// Module Structure Definitions
//

//
// Module Function Prototypes
//

//
// Module Variables
//
static int GosubFlag =  0;
static int OnErrorFlag = 0;

static int IOChPrinted;		// Has channel number been output
static Node* IOChannel;		// IO Channel to use (NULL if default)
static Node* IOUsing;		// Print using format

/*******************************************************************************

Member Function:

	Node()

Title:

	Constructs a node.

Description:

	This function creates a new node.

*******************************************************************************/

int Node::Level = 0;

Node::Node(int TypeValue, char* xTextValue, int Include)
{
	Type = TypeValue;
	TextValue = NULL;
	Down = NULL;
	Tree[0] = NULL;
	Tree[1] = NULL;
	Tree[2] = NULL;
	Tree[3] = NULL;
	FromInclude = Include;

	//
	// Allocate room for text, and copy over.
	//
	if (xTextValue != NULL)
	{
		char *ThisText = new char[strlen(xTextValue) + 1];
		assert(ThisText != NULL);
		strcpy(ThisText, xTextValue);

		if ((Type != BAS_V_TEXTSTRING) && (Type != BAS_S_REMARK) &&
			(Type != BAS_V_INTEGER))
		{
			strupr(ThisText);
		}
		TextValue = ThisText;
	}
}

/*******************************************************************************

Member Function:

	~Node()

Title:

	Destructs a node, and all of it's children.

Description:

	This function frees an existing node, and all of its children.

*******************************************************************************/

//
// Destructor
//
Node::~Node()
{
	// Lose all parameters

	delete Tree[0];
	delete Tree[1];
	delete Tree[2];
	delete Tree[3];

	// Lose text value

	delete TextValue;
}

/*******************************************************************************


	Link()

Title:

	Link child nodes into parent node.

Description:

	This function will link several child nodes into a parent node.

Return Value:

	Returns a pointer to the parent node (passed as the first
	parameter)

Bugs:

*******************************************************************************/

Node *Node::Link
(
	Node *Node0,
	Node *Node1,
	Node *Node2,
	Node *Node3
)
{
	//
	// Attach them together
	//
	if (Node0 != NULL)
	{
		assert(Tree[0] == NULL);
		assert(Node0 != this);
		Tree[0] = Node0;
	}
	if (Node1 != NULL)
	{
		assert(Tree[1] == NULL);
		assert(Node1 != this);
		Tree[1] = Node1;
	}
	if (Node2 != NULL)
	{
		assert(Tree[2] == NULL);
		assert(Node2 != this);
		Tree[2] = Node2;
	}
	if (Node3 != NULL)
	{
		assert(Tree[3] == NULL);
		assert(Node3 != this);
		Tree[3] = Node3;
	}

	//
	// Return main node back
	//
	return(this);
}

/*******************************************************************************

Member Function:

	DownLink()

Title:

	Link two nodes together at same level.

Description:

	This function will link two parent nodes together at the same
	tree level. If one or the other node doesn't exist, it will
	make whichever node exists the main node.

Return Value:

	Returns the top node.
Bugs:

*******************************************************************************/

//
// Link several nodes together through the Nodes
//
Node *Node::DownLink
(
	Node *Node0
)
{
	//
	// Don't link if nothing to link to
	//
	if (this == NULL)
	{
		return(Node0);
	}

	//
	// Skip out if link node is null
	//
	if (Node0 == NULL)
	{
		return this;
	}

	assert(Node0 != this);

	//
	// If we aren't at the top of the stack, scan to the end
	//
	Node* LookDown = this;
	while (LookDown->Down != NULL)
	{
		LookDown = LookDown->Down;
	}

	//
	// Attach them together
	//
	LookDown->Down = Node0;

	//
	// Return main node back
	//
	return(this);
}

/*******************************************************************************

Global Function:

	PrintTree()

Title:

	Prints the node tree.

Description:

	This function will print out a simple picture of the node
	and its relationship to it's children.

Bugs:

	Only displays to stdout. Ugly display.

*******************************************************************************/

//
// Print out tree in a tree structure
//
void Node::PrintTree(char* Level, int Flag, char Node)
{
	//
	// Not really a node
	//
	assert(this != NULL);

	//
	// Local variables
	//
	char* DownLevel;

	//
	// Make sure we have an active node
	//
	if ((Down != NULL) || Flag)
	{
		DownLevel = " | ";
	}
	else
	{
		DownLevel = "   ";
	}

	char* NextLevel = new char[strlen(Level) + 4];
	strcpy(NextLevel, Level);
	strcat(NextLevel, DownLevel);

	//
	// Print bars for this level
	//
	cout << Level << " " << Node << "-" << Type;

	if (TextValue != NULL)
	{
		cout << " Text: " << TextValue;
	}

	//
	// Output variable type
	//
	cout << " <" << DumpVarType(GetNodeVarType()) << ">";

	cout << endl;

	//
	// Print Parameter Levels
	//
	if (Tree[0] != NULL)
	{
		Tree[0]->PrintTree(NextLevel,
			(Tree[1] != NULL) | (Tree[2] != NULL) |
			(Tree[3] != NULL), '0');
	}

	if (Tree[1] != NULL)
	{
		Tree[1]->PrintTree(NextLevel,
			(Tree[2] != NULL) | (Tree[3] != NULL), '1');
	}
	if (Tree[2] != NULL)
	{
		Tree[2]->PrintTree(NextLevel, (Tree[3] != NULL), '2');
	}

	if (Tree[3] != NULL)
	{
		Tree[3]->PrintTree(NextLevel, 0, '3');
	}

	//
	// Link down to next parent if possible
	//
	if (Down != NULL)
	{
		Down->PrintTree(Level, 0, 'D');
	}
}

/*******************************************************************************

Member Function:

	VariableScan()

Title:

	Search block for variable names

Description:

	Scans a block of code for variable names.
	This must be done before scanning for the loops.
	Call this funstion to process a block of statements.

Bugs:

*******************************************************************************/

void Node::VariableScan(int InDefineFlag)
{
	Node *ThisNode;

	assert(this != NULL);

	//
	// Scan this statement
	//
	VariableScanOne(InDefineFlag);

	//
	// Scan Next Statement(s)
	//
	//	I'm doing the scan here instead of having each node
	//	call scan on it's own Dowm because the recursion becomes
	//	a real memory killer on large programs.
	//
	ThisNode = Down;

	while (ThisNode != NULL)
	{
		ThisNode->VariableScanOne(InDefineFlag);
		ThisNode = ThisNode->Down;
	}
}

/*******************************************************************************

Member Function:

	VariableScanOne()

Title:

	Search statement for variable names.

Description:

	Scans a block of code for variable names.
	This must be done before scanning for the loops.
	Call this function to process a single statement.

Bugs:

*******************************************************************************/

void Node::VariableScanOne(int InDefineFlag)
{
	VariableStruct *ThisVar;
	int ThisType = 0;

	assert(this != NULL);

	//
	// Look at current node type
	//
	switch (Type)
	{
	case BAS_N_EXTERNALCONSTANT:
		assert(Tree[2] != NULL);
		if (Tree[1] != NULL)
		{
			ThisType = Tree[1]->ScanType();
		}
		else
		{
			ThisType = 0;
		}

		Tree[2]->ScanVarList(ThisType, VARCLASS_CONST, TRUE);

		break;

	case BAS_N_EXTERNALFUNCTION:
		assert(Tree[1] != NULL);
		ThisType = Tree[1]->ScanType();
		Tree[2]->ScanVarList(ThisType, VARCLASS_FUNC, TRUE);
		break;

	case BAS_N_EXTERNALSUB:
		Tree[1]->ScanVarList(VARTYPE_VOID, VARCLASS_FUNC, TRUE);
		break;

	case BAS_S_CALL:
		Tree[0]->ScanVarList(VARTYPE_VOID, VARCLASS_FUNC, FALSE);
		break;

	case BAS_S_COMMON:
		assert(Tree[0] != NULL);

		ThisVar = new VariableStruct(Tree[0]->TextValue,
			VARTYPE_STRUCT, VARCLASS_NONE, TRUE);
		Variables->Append(ThisVar);
		Variables->Fixup();

		break;

	case BAS_S_DATA:
		break;

	case BAS_S_DECLARE:
		assert(Tree[2] != NULL);
		if (Tree[1] != NULL)
		{
			ThisType = Tree[1]->ScanType();
		}
		else
		{
			ThisType = VARTYPE_REAL;
		}

		Tree[2]->ScanVarList(ThisType, VARCLASS_NONE, TRUE);

		break;

	case BAS_S_DIM:
		assert(Tree[0] != NULL);
		if (Tree[1] != NULL)
		{
			Tree[1]->VariableScanOne(1);
		}
		Tree[0]->ScanVarList(VARTYPE_NONE, VARCLASS_ARRAY, TRUE);
		break;

	case BAS_S_EXTERNAL:
		assert(Tree[2] != NULL);
		if (Tree[1] != NULL)
		{
			ThisType = Tree[1]->ScanType();
		}
		else
		{
			ThisType = 0;
		}

		Tree[2]->ScanVarList(ThisType, VARCLASS_NONE, TRUE);

		break;

	case BAS_S_MAP:

		ScanMap();
		break;

	case BAS_V_FUNCTION:
		//
		// Built in function
		//
		ThisVar = Variables->Lookup(TextValue, this);
		if (ThisVar == NULL)
		{
			//
			// Create variable
			//
			ThisVar = new VariableStruct(TextValue,
				VARTYPE_VOID, VARCLASS_FUNC, InDefineFlag);
			assert(ThisVar != NULL);

			//
			// Add to variable table
			//
			Variables->Append(ThisVar, 1);		// Forced global
		}
		else
		{
			ThisVar->SetClass(VARCLASS_FUNC);
		}
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScan(InDefineFlag);
		}
		if (Tree[3] != NULL)
		{
			Tree[3]->VariableScan(0);
		}
		break;

	case BAS_S_FUNCTION:
	case BAS_S_MAINFUNCTION:
		//
		// Function Definition
		//
		assert(Tree[2] != NULL);

		if (Tree[0] != NULL)
		{
			ThisType = Tree[0]->ScanType();
		}
		else
		{
			ThisType = VARTYPE_VOID;
		}

		ThisVar = Variables->Lookup(Tree[2]->TextValue, this);
		if (ThisVar == NULL)
		{
			//
			// Create variable
			//
			ThisVar = new VariableStruct(Tree[2]->TextValue,
				ThisType, VARCLASS_FUNC, InDefineFlag);
			assert(ThisVar != NULL);

			//
			// Add to variable table
			//
			Variables->Append(ThisVar, 1);		// Forced global
		}
		else
		{
			ThisVar->SetClass(VARCLASS_FUNC);
			ThisVar->Type = ThisType;
		}

		//
		// Scan calling parameters
		//
		if (Tree[2]->Tree[0] != NULL)
		{
			Tree[2]->Tree[0]->VariableScan(1);
		}

		//
		// Scan source code
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->VariableScan(0);
		}

		break;

	case BAS_S_DEF:
		//
		// Function Definition
		//
		assert(Tree[2] != NULL);

		//
		// Try to determint data type of this def
		//
		if (Tree[0] != NULL)
		{
			//
			// DEF type name
			//
			ThisType = Tree[0]->ScanType();
		}
		else
		{
			//
			// DEF name
			//
			switch(Tree[2]->TextValue[strlen(Tree[2]->TextValue) - 1])
			{
			case '$':
				ThisType = VARTYPE_DYNSTR;
				break;

			case '%':
				ThisType = VARTYPE_INTEGER;
				break;

			default:
				ThisType = VARTYPE_REAL;
				break;
			}
		}

		ThisVar = Variables->Lookup(Tree[2]->TextValue, this);
		if (ThisVar == NULL)
		{
			//
			// Create variable
			//
			ThisVar = new VariableStruct(Tree[2]->TextValue,
				ThisType, VARCLASS_FUNC, InDefineFlag);
			assert(ThisVar != NULL);

			//
			// Add to variable table
			//
			Variables->Append(ThisVar, 1);		// Forced global
		}
		else
		{
			ThisVar->SetClass(VARCLASS_FUNC);
			ThisVar->Type = ThisType;
		}

		//
		// Scan calling parameters
		//
		if (Tree[2]->Tree[0] != NULL)
		{
			Tree[2]->Tree[0]->VariableScan(1);
		}

		//
		// Scan source code
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->VariableScan(0);
		}

		break;

	case BAS_V_LABEL:
		ThisVar = Variables->Lookup(TextValue, 0);
		if (ThisVar == NULL)
		{
			int SeekColon = strlen(TextValue);
			if (TextValue[SeekColon - 1] == ':')
			{
				TextValue[SeekColon-1] = '\0';
			}
			ThisVar = new VariableStruct(TextValue,
				VARTYPE_LABEL, VARCLASS_LABEL, TRUE);
			assert(ThisVar != NULL);
			Variables->Append(ThisVar);
		}
		break;

	case BAS_V_USELABEL:
		ThisVar = Variables->Lookup(TextValue, 0);
		if (ThisVar == NULL)
		{
			ThisVar = new VariableStruct(TextValue,
				VARTYPE_LABEL, VARCLASS_LABEL, TRUE);
			assert(ThisVar != NULL);
			Variables->Append(ThisVar);
		}
		ThisVar->EverUsed++;
		break;

	case BAS_V_NAME:
		Variables->Define(this, FALSE, InDefineFlag);

		//
		// Look at possible array reference
		//
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScanOne(InDefineFlag);
		}

		if (Tree[2] != NULL)
		{
			Tree[2]->VariableScanOne(InDefineFlag);
		}

		break;

	case BAS_S_RECORD:
		assert(Tree[0] != NULL);
		assert(Tree[3] != NULL);
		Tree[0]->VariableScanOne(1);
		Tree[3]->VariableScan(1);
		break;

	case BAS_N_STRUCTNAME:

		//
		// Force structer variable to be usable under C++
		//
		strlwr(TextValue);

		//
		// Look at possible array reference
		//
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScanOne(InDefineFlag);
		}

		if (Tree[2] != NULL)
		{
			Tree[2]->VariableScanOne(InDefineFlag);
		}

		break;

	case BAS_X_STRREF:

		assert (Tree[0] != NULL);
		assert (Tree[1] != NULL);

		Tree[0]->VariableScanOne(InDefineFlag);
		Tree[1]->VariableScanOne(InDefineFlag);

		break;

	case BAS_S_GOSUB:
		GosubFlag = 1;
		assert(Tree[0] != NULL);
		Tree[0]->VariableScanOne(InDefineFlag);
		break;

	case BAS_N_ONERROR:
		OnErrorFlag = 1;
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScanOne(InDefineFlag);
		}
		break;

	case BAS_N_ONGOSUB:
		GosubFlag = 1;
		assert(Tree[0] != NULL);
		assert(Tree[1] != NULL);
		Tree[0]->VariableScanOne(InDefineFlag);
		Tree[1]->VariableScanOne(InDefineFlag);
		break;

	case BAS_S_RETURN:
		GosubFlag = 1;
		break;

	case BAS_S_RESUME:
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScanOne(InDefineFlag);
		}
		OnErrorFlag = 1;
		break;

	//
	// This probibly should be in parse.y.
	//
	case '+':
		//
		// Process lower levels first
		//
		Tree[0]->VariableScanOne(InDefineFlag);
		Tree[1]->VariableScanOne(InDefineFlag);

		//
		// Let's try and merge text strings together
		//
		if (Tree[0]->IsReallyString() && Tree[1]->IsReallyString())
		{
			//
			// Create merged string (Using the colon
			// concatination options)
			//
			assert(Tree[0]->TextValue != NULL);
			assert(Tree[1]->TextValue != NULL);
			char* Merger = new char[strlen(Tree[0]->TextValue) +
				strlen(Tree[1]->TextValue) + 1];
			assert(Merger != NULL);

			strcpy(Merger, Tree[0]->TextValue);
			strcat(Merger, Tree[1]->TextValue);

			//
			// Change the root node to the concatinated string
			// type.
			//
			Type = BAS_V_TEXTSTRING;
			delete Tree[0];
			Tree[0] = NULL;
			delete Tree[1];
			Tree[1] = NULL;
			delete TextValue;
			TextValue = Merger;
		}
		break;

	default:
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScan(InDefineFlag);
		}

		if (Tree[1] != NULL)
		{
			Tree[1]->VariableScan(InDefineFlag);
		}

		if (Tree[2] != NULL)
		{
			Tree[2]->VariableScan(InDefineFlag);
		}

		if (Tree[3] != NULL)
		{
			Tree[3]->VariableScan(InDefineFlag);
		}
		break;
	}
}


/*******************************************************************************

Member Function:

	Node::ScanVarList()

Title:

	Scans a (possible) list for the variable names of the
	current type.

Description:

	Scans a block of code for the variable type.

Bugs:

*******************************************************************************/

void Node::ScanVarList(int ThisType, int ThisClass, int Status)
{
	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:
	case BAS_N_ASSIGNLIST:

		Tree[0]->ScanVarList(ThisType, ThisClass, Status);
		Tree[1]->ScanVarList(ThisType, ThisClass, Status);
		break;

	case '=':
	case BAS_N_ASSIGN:

		Tree[0]->ScanVarList(ThisType, ThisClass, Status);
		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		ThisType = this->ScanType();
		Tree[2]->ScanVarList(ThisType, ThisClass, Status);
		break;

	default:
		VariableStruct* ThisVar = Variables->Lookup(TextValue, Tree[0]);
		if (ThisVar == NULL)
		{
			int FinalType = ThisType;
			if (FinalType == VARTYPE_NONE)
			{
				FinalType = GuessVarType(TextValue);
			}
			ThisVar = new VariableStruct(TextValue,
				FinalType, ThisClass, Status);
			assert(ThisVar != NULL);
			Variables->Append(ThisVar);
		}
	}
}


/*******************************************************************************

Member Function:

	Node::ScanType()

Title:

	Search block for type definion

Description:

	Scans a block of code for the variable type.

Bugs:

*******************************************************************************/

int Node::ScanType()
{
	//
	// Don't even try it is we don't know what's happening
	//
	assert(this != NULL);

	//
	// Try to figure out the variable type
	//
	switch (Type)
	{
	case BAS_S_REAL:
		return(RealType);

	case BAS_S_DOUBLE:
		return(VARTYPE_DOUBLE);

	case BAS_S_BYTE:
		return(VARTYPE_BYTE);

	case BAS_S_GFLOAT:
		return(VARTYPE_DOUBLE);

	case BAS_S_HFLOAT:
		return(VARTYPE_DOUBLE);

	case BAS_S_INTEGER:
		return(VARTYPE_INTEGER);

	case BAS_S_LONG:
		return(VARTYPE_LONG);

	case BAS_S_RFA:
		return(VARTYPE_RFA);

	case BAS_S_STRING:
		return(VARTYPE_DYNSTR);

	case BAS_S_WORD:
		return(VARTYPE_WORD);

	case BAS_S_DECIMAL:
		return(VARTYPE_DECIMAL);
	}

	return(0);
}

/*******************************************************************************

Member Function:

	Node::Output()

Title:

	Main function to output translated code.

Description:

	This function outputs the translated code.

Bugs:

	Ugly Code.

*******************************************************************************/

void Node::Output(ostream& os)
{
	//
	// Handle Include Files
	//
	if (NeedIostreamH != 0)
	{
		os << "#include <iostream.h>" << endl;
		NeedIostreamH = 0;
	}

	os << "#include <stdlib.h>" << endl;
	NeedStdlibH = 0;

//	if (NeedUnistd != 0)
//	{
		os << "#include <unistd.h>" << endl;
		NeedUnistd = 0;
//	}

	if ((NeedStringH != 0) | (NeedBasicFun != 0))
	{
		os << "#include <string.h>" << endl;
		NeedStringH = 0;
	}

	if (NeedMathH != 0)
	{
		os << "#include <math.h>" << endl;
		NeedMathH = 0;
	}

	if (NeedErrorH != 0)
	{
		os << "#include <errno.h>" << endl;
		NeedErrorH = 0;
	}

	if (NeedTimeH != 0)
	{
		os << "#include <time.h>" << endl;
		NeedTimeH = 0;
	}

	os << "#include \"basicfun.h\"" << endl;
	os << "#include \"bstring.h\"" << endl;
	if (NeedChannel != 0)
	{
		os << "#include \"basicchannel.h\"" << endl;
		os << "BChannel BasicChannel[MaxChannel + 1];" << endl;
		NeedChannel = 0;
	}
	NeedBasicFun = 0;

	if (NeedDataList != 0)
	{
		os << "#include \"datalist.h\"" << endl;
		NeedDataList = 0;
	}

	if (NeedPuse != 0)
	{
		os << "#include <pusing.h>" << endl;
	}

	os << endl;

	if (NeedRFA != 0)
	{
		os << "typedef char RmsRfa[6];" << endl;
		NeedRFA = 0;
	}

	if (NeedPuse != 0)
	{
		os << "static PUsing PUse;\t// Needed for print using" << endl;
		NeedPuse = 0;
	}

	os << endl;

	//
	// Output the variables
	//
	Variables->OutputDef(os, Level);

	//
	// Dump out prototypes
	//
	OutputPrototypes(os);

	//
	// Output code
	//
	Level = 0;
	OutputCode(os);
}


/*******************************************************************************

Local Function:

	Node::OutputCode()

Title:

	Output one block of code.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

void Node::OutputCode(ostream& os)
{
	assert(this != NULL);

	//
	// Local Variables
	//
	Node *Program = this;

	//
	// Starting at the top node, we scan down handling each
	// level 0 node type as necessary.
	//
	// Could do this by calling OutputCode recursively for the Down
	// level instead of splitting into OutputCode and OutputCodeOne,
	// but that would probibly consume too much stack space.
	//
	while (Program != NULL)
	{
		//
		// Handle Statement. Don't output code from included files.
		//
		if ((Program->FromInclude == 0) || (CompileFlag != 0))
		{
			os << Program;
		}

		//
		// Get next node
		//
		Program = Program->Down;
	}
}


/*******************************************************************************

Local Function:

	Node::OutputBlock()

Title:

	Output one block of code.

Description:

	This function will output a block of code (if it exists)
	surrounded by parentheses, otherwise it will output a semicolon.

Bugs:


*******************************************************************************/

void Node::OutputBlock(ostream& os)
{
	//
	// If this is a null bit of code, just output a semicolon
	//
	if (this == NULL)
	{
		Indent(os);
		os << ";" << endl;
	}
	else
	{
		//
		// Start the block
		//
		Indent(os);
		os << "{" << endl;

		Level++;

		//
		// Local Variables
		//
		Node *Program = this;

		//
		// Starting at the top node, we scan down handling each
		// level 0 node type as necessary.
		//
		// Could do this by calling OutputCode recursively for the Down
		// level instead of splitting into OutputCode and OutputCodeOne,
		// but that would probibly consume too much stack space.
		//
		while (Program != NULL)
		{
			//
			// Handle Statement
			//
			if ((Program->FromInclude == 0) || (CompileFlag != 0))
			{
				os << Program;
			}

			//
			// Get next node
			//
			Program = Program->Down;
		}

		//
		// End the block
		//
		Level--;
		Indent(os);
		os << "}" << endl;
	}
}


/*******************************************************************************

Local Function:

	Node::NoParen()

Title:

	Return a node losing any parentheses around the whole group.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

Node& Node::NoParen()
{
	if (Type == '(')
	{
		return Tree[0]->NoParen();
	}
	else
	{
		return *this;
	}
}

/*******************************************************************************

Local Function:

	Node::OutputPrototypes()

Title:

	Output function Prototypes.

Description:

	This function scans through the basic blocks and dumps out
	function prototypes for all functions defined therin.

Bugs:


*******************************************************************************/

void Node::OutputPrototypes(ostream& os)
{
	Node* ThisNode = this;

	/*
	 * Output title for prototype section
	 */
	os << endl << "//" << endl << "// Function Prototypes" << endl <<
		"//" << endl;

	//
	// Go through the level 0 functions
	//
	while(ThisNode != NULL)
	{

		switch(ThisNode->Type)
		{
		//
		// Functions
		//
		case BAS_S_FUNCTION:
		case BAS_S_DEF:

			//
			// Empty variable list
			//
			Variables->ClearLocal();

			//
			// Scan the call parameters
			//
			ThisNode->VariableScanOne(1);
			Variables->Fixup();

			//
			// Output function name
			//
			Indent(os);
			ThisNode->Tree[2]->OutputDefinition(os,
				ThisNode->Tree[0], ThisNode->Type);
			os << ";" << endl;

			break;

		//
		// Subs
		//
		case BAS_S_SUB:

			//
			// Empty variable list
			//
			Variables->ClearLocal();

			//
			// Scan the call parameters
			//
			ThisNode->VariableScanOne(1);
			Variables->Fixup();

			//
			// Output function name
			//
			Indent(os);
			ThisNode->Tree[2]->OutputDefinition(os,
				ThisNode->Tree[0], ThisNode->Type);
			os << ";" << endl;
			break;
		}

		ThisNode = ThisNode->Down;
	}

	os << endl;
}

/*******************************************************************************

Local Function:

	Node::OutputCodeOne()

Title:

	Output one block of code.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

void Node::OutputCodeOne(ostream& os)
{
	VariableStruct *ThisVar;
	char *ThisPart = 0;
	Node* LookDown;

	assert(this != NULL);

	//
	// Handle this node type
	//
	switch (Type)
	{
	//
	// Assignment statement
	//
	case '#':
		os << "BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "]";
		break;

	case '=':
		Tree[0]->OutputBstringText(os);
		os << " == " << Tree[1];
		break;

	case '+':
	case '-':
	case '*':
	case '/':
	case '>':
	case '<':
		Tree[0]->OutputBstringText(os);
		os << " " << TextValue << " " << Tree[1];
		break;

	case '(':
		//
		// Lose unecessary additional levels of pren's
		//
		os << "(" << Tree[0]->NoParen() << ")";
		break;

	case '^':
		os << "pow(" << Tree[0]->NoParen() << ", " <<
			Tree[1]->NoParen() << ")";
		break;

	case BAS_X_LAND:
		os << Tree[0] << " && " << Tree[1];
		break;

	case BAS_S_AND:
		os << Tree[0] << " & " << Tree[1];
		break;

	case BAS_N_ASSIGN:
		Indent(os);
		if (Tree[0] != NULL)
		{
			//
			// Function return values are stashed in the
			// 'Result' variable.
			//
			if (Tree[0]->Type == BAS_V_FUNCTION)
			{
				os << "Result = " << Tree[1] << ";" << endl;
			}
			else
			{
				//
				// A simple assignment statement
				//
				os << Tree[0] << " = " << Tree[1] <<
					";" << endl;
			}
		}
		else
		{
			os << "Result = " << Tree[1] << ";" << endl;
		}
		break;

	case BAS_N_ASSIGNLIST:
		assert(Tree[0] != NULL);
		assert(Tree[1] != NULL);

		if (Tree[0]->Type == BAS_V_FUNCTION)
		{
			os << "Result";
		}
		else
		{
			os << Tree[0];
		}

		os << " = ";

		if (Tree[1]->Type == BAS_V_FUNCTION)
		{
			os << "Result";
		}
		else
		{
			os << Tree[1];
		}
		break;

	case BAS_S_BY:
		switch(Tree[1]->Type)
		{
		case BAS_S_DESC:
			os << "$descr(" << Tree[0] << ")";
			break;
		case BAS_S_REF:
			os << "&" << Tree[0];
			break;
		default:
			os << Tree[0];
			break;
		}
		break;

	case BAS_S_CALL:
		Indent(os);
		os << Tree[0] << ";" << endl;
		break;

	case BAS_S_CASE:
		if (Tree[0] != NULL)
		{
			Tree[0]->OutputCaseLabel(os);
			os << endl;
		}
		else
		{
			Indent(os);
			os << "case:\t// Empty case label" <<
				endl << endl;
		}

		//
		// Output any info under case statement
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->OutputCode(os);
		}
		Indent(os);
		os << "break;" << endl << endl;
		break;

	case BAS_N_CASEELSE:
		Level--;
		Indent(os);
		os << "default:" << endl;
		Level++;

		//
		// Output any info under case statement
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->OutputCode(os);
		}
		Indent(os);
		os << "break;" << endl << endl;
		break;

	case BAS_P_CDD:
		os << "#dictionary " << Tree[0] << endl;
		break;

	case BAS_S_CHAIN:
		Indent(os);
		os << "BasicChain(" << Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_S_CLOSE:
		Indent(os);
		os << "BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Close();" << endl;
		break;

	case BAS_S_COMMON:
		OutputMap(os);
		break;

	case BAS_S_DATA:
		Indent(os);
		os << "//" << endl;
		Indent(os);
		os << "// Data Statement" << endl;
		Indent(os);
		os << "//" << endl;
		Indent(os);
		os << "static const char* DataValue[] = {" << endl;
		Level++;
		OutputData(os);
		os << ", NULL};" << endl;
		Level--;
		Indent(os);
		os << "DataListClass DataList(DataValue);" << endl;
		break;

	case BAS_S_DECLARE:
		Tree[2]->OutputDefinitionList(os, Tree[1], Type);
		break;

	case BAS_S_DELETE:
		Indent(os);
		os << "BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Delete();" << endl;
		break;

	case BAS_S_DIM:
		assert(Tree[0] != NULL);
		if (Tree[1] == NULL)
		{
			Tree[0]->OutputDefinitionList(os, this, Type);
		}
		else
		{
			//
			// Set up a place marker for now
			//
			Indent(os);
			os << "VirtualArray(" << Tree[1] << ")" << endl;
			Indent(os);
			os << "{" << endl;
			Level++;
			Tree[0]->OutputDefinitionList(os, this, Type);
			Level--;
			Indent(os);
			os << "};" << endl;
		}
		break;

	case BAS_S_ELSE:
		break;

	case BAS_S_END:
	case BAS_S_FNEND:
		break;

	case BAS_S_EQV:
	case BAS_S_IMP:
		os << "Basic" << TextValue << "(" << Tree[0] << ", " <<
			Tree[1] << ")";
		break;

	case BAS_S_ERL:
		os << "ErrLine";
		break;

	case BAS_S_ERR:
		os << "ErrNum";
		break;

	case BAS_S_EXTERNAL:
		Tree[2]->OutputDefinitionList(os, Tree[1], Type);
		break;

	case BAS_N_EXTERNALCONSTANT:
		Tree[2]->OutputDefinitionList(os, Tree[1], Type);
		break;

	case BAS_N_EXTERNALFUNCTION:
		Tree[2]->OutputDefinitionList(os, Tree[1], Type);
		break;

	case BAS_N_EXTERNALSUB:
		Tree[2]->OutputDefinitionList(os, Tree[1], Type);
		break;

	case BAS_S_EXIT:
		Indent(os);
		os <<  "return;" << endl;
		break;

	case BAS_S_FIELD:
		Indent(os);
		os << "// Field Statement" << endl;
		Indent(os);
		os << "{" << endl;
		Level++;
		Indent(os);
		os << "char* FieldBase = BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].BufferLoc();" << endl;
		Tree[1]->OutputField(os);
		Level--;
		Indent(os);
		os << "}" << endl;
		break;

	case BAS_S_FIND:
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputGetPutOptions(Tree[0], os);
		}
		Indent(os);
		os << "BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Find()" << endl;
		break;

	case BAS_S_FOR:
		Indent(os);
		os << "for (";

		//
		// from
		//
		os << Tree[0] << "; ";

		//
		// to
		//
		//	We have to try to look at the step and guess if
		//	it is a negitive step value. If we can't tell,
		//	assume positive.
		//
		if ((Tree[2] != NULL) &&
			(Tree[2]->Type == BAS_N_UMINUS))
		{
			os << Tree[0]->Tree[0] << " >= " << Tree[1] << "; ";
		}
		else
		{
			os << Tree[0]->Tree[0] << " <= " << Tree[1] << "; ";
		}

		//
		// step
		//
		if (Tree[2] != NULL)
		{
			//
			// Loop going backwards?
			//
			if (Tree[2]->Type == BAS_N_UMINUS)
			{
				os << Tree[0]->Tree[0] << " -= " <<
					Tree[2]->Tree[0];
			}
			else
			{
				os << Tree[0]->Tree[0] << " += " << Tree[2];
			}
		}
		else
		{
			os << Tree[0]->Tree[0] << "++";
		}

		os << ")" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_N_FORASSIGN:
		os << Tree[0] << " = " << Tree[1];
		break;

	case BAS_N_FORUNTIL:
		Indent(os);

		//
		// from
		//
		os << "for (" << Tree[0] << "; ";

		//
		// until
		//
		os << "!(" << Tree[0]->NoParen() << ")" << "; ";

		//
		// step
		//
		os << Tree[0]->Tree[0] << "++)" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_N_FORWHILE:
		Indent(os);

		//
		// from
		//
		os << "for (" << Tree[0] << "; ";

		//
		// until
		//
		os << Tree[0]->Tree[0] << "==" << Tree[1] << "; ";

		//
		// step
		//
		os << Tree[0]->Tree[0] << "++)" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_FUNCTION:
	case BAS_S_MAINFUNCTION:
	case BAS_S_DEF:

		//
		// Empty variable list
		//
		assert(Tree[2] != 0);
		Variables->ClearLocal();

		//
		// Scan the call parameters
		//
		this->VariableScanOne(1);
		Variables->Fixup();

		//
		// Dump out variable names if user asked for it
		//
		if (VariableDump)
		{
			Variables->Dump();
		}

		//
		// Output function name
		//
		os << endl;
		Indent(os);
		if (Type == BAS_S_MAINFUNCTION)
		{
			os << "int main(int argc, char** argv)";
		}
		else
		{
			Tree[2]->OutputDefinition(os, Tree[0], Type);
		}
		os << endl;

		//
		// Braces starting function
		//
		Indent(os);
		os << "{" << endl;
		Level++;

		//
		// Variable to hold result code
		//
		//	NOTE: Need to generate actual data type here.
		//
		if (Type != BAS_S_MAINFUNCTION)
		{
			Indent(os);
			Tree[2]->OutputDefinitionCopy(os, Tree[0], Type,
				"Result");
			os << ";" << endl;
		}

		//
		// Output local variables
		//
		if (Tree[3] != 0)
		{
			Variables->OutputDef(os, Level);
			os << endl;

			if (GosubFlag != 0)
			{
				Indent(os);
				os << "BStack(20);" << endl;
				GosubFlag = 0;
			}

			if (OnErrorFlag != 0)
			{
				Indent(os);
				os << "OnErrorStack;" << endl;
				OnErrorFlag = 0;
			}

			Tree[3]->OutputCode(os);
		}

		//
		// Give back a result code
		//
		if (Type != BAS_S_MAINFUNCTION)
		{
			Indent(os);
			os << "return Result;" << endl;
		}

		Level--;
		Indent(os);
		os << "}" << endl;

		break;

	case BAS_V_FUNCTION:
		//
		// Look for variable in table
		//
		ThisVar = Variables->Lookup(TextValue, Tree[0]);

		//
		// Print this variable
		//
		if (ThisVar != NULL)
		{
			os << ThisVar->GetName();
		}
		else
		{
			os << TextValue;
		}

		if (Tree[0] != NULL)
		{
			os << "(" << Tree[0]->NoParen() << ")";
		}
		else
		{
			os << "()";
		}
		break;

	case BAS_S_FUNCTIONEXIT:
		Indent(os);
		os << "return Result;" << endl;
		break;

	case BAS_X_GE:
		os << Tree[0] << " >= " << Tree[1];
		break;

	case BAS_S_GET:
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputGetPutOptions(Tree[0], os);
		}
		Indent(os);
		os << "BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Get();" << endl;
		//
		// Options
		//
//		if (Tree[1] != NULL)
//		{
//			os << Tree[1]->NoParen();
//		}
//		os << ");" << endl;
		break;

	case BAS_S_GOSUB:
		//
		// Yes, I know, C doesn't have a GOSUB, but this puts
		// out code that can be manually edited. Without this,
		// there is currently no way to find the subroutines.
		//
		Indent(os);
		os << "BGosub(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_GOTO:
		Indent(os);
		os << "goto " << Tree[0] << ";" << endl;
		break;

	case BAS_S_IF:
		Indent(os);
		assert(Tree[0] != NULL);

		os << "if (" << Tree[0]->NoParen() << ")" << endl;

		Tree[3]->OutputBlock(os);

		if (Tree[2] != NULL)
		{
			Indent(os);
			os << "else" << endl;

			Tree[2]->OutputBlock(os);
		}
		break;

	case BAS_P_INCLUDE:
		//
		// Don't output '#include' if we are also outputting the
		// code that was included.
		//
		if (CompileFlag == 0)
		{
			os << "#include ";
		}
		else
		{
			os << "// #include ";
		}
		os << Tree[0];

		//
		// How should this really be handled?
		//
		if (Tree[1] != NULL)
		{
			os << " %library " << Tree[1];
		}
		os << endl;

		break;

	case BAS_S_INPUT:
		OutputInput(os, 1);
		break;

	case BAS_V_INT:
		OutputInteger(os);
		break;

	case BAS_V_INTEGER:
	case BAS_V_FLOAT:
		OutputInteger(os);
		break;

	case BAS_S_ITERATE:
		Indent(os);
		os << "continue;" << endl;
		break;

	case BAS_S_KILL:
		Indent(os);
		os << "unlink(" << Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_V_LABEL:
		ThisVar = Variables->Lookup(TextValue, 0);
		if (ThisVar != NULL)
		{
			if ((ThisVar->EverUsed != 0) || (KeepAllLines == TRUE))
			{
				os << ThisVar->GetName() << ":;" << endl;
			}
		}
		else
		{
			os << TextValue << ":;" << endl;
		}
		break;

	case BAS_X_LE:
		os << Tree[0] << " <= " << Tree[1];
		break;

	case BAS_S_LET:
		Indent(os);
		os << Tree[0] << ";" << endl;
		break;

	case BAS_S_LINPUT:
		OutputInput(os, 2);
		break;

	case BAS_N_LIST:
		if (Tree[0] == NULL)
		{
			os << "0";
		}
		else
		{
			os << Tree[0];
		}

		os << ", ";

		if (Tree[1] == NULL)
		{
			os << "0";
		}
		else
		{
			os << Tree[1];
		}
		break;

	case BAS_S_LSET:
		assert(Tree[0] != NULL);
		assert(Tree[0]->Type == BAS_N_ASSIGN);
		assert(Tree[0]->Tree[0] != NULL);
		assert(Tree[0]->Tree[1] != NULL);

		Indent(os);
		os << Tree[0]->Tree[0] << ".Lset(" <<
			Tree[0]->Tree[1] << ");" << endl;
		break;

	case BAS_S_MAP:
		OutputMap(os);
		break;

	case BAS_V_NAME:
		OutputVarName(os);
		break;

	case BAS_X_NEQ:
		os << Tree[0] << " != " << Tree[1];
		break;

	case BAS_S_NOT:
		os << "~" << Tree[0];
		break;

	case BAS_X_LNOT:
		os << "!" << Tree[0];
		break;

	case BAS_N_NULL:
		break;

	case BAS_N_ONERROR:
		Indent(os);
		if (atoi(Tree[0]->TextValue) == 0)
		{
			os << "OnErrorZero;" << endl;
		}
		else
		{
			os << "OnErrorGoto(" << Tree[0] << ");" << endl;
		}
		break;

	case BAS_N_ONGOTO:
		Indent(os);
		os << "switch(" << Tree[0]->NoParen() << ")" << endl;
		Indent(os);
		os << "{" << endl;
		Tree[1]->OutputOngo(os, "goto", 1);
		Indent(os);
		os << "}" << endl;
		break;

	case BAS_N_ONGOSUB:
		Indent(os);
		os << "switch(" << Tree[0]->NoParen() << ")" << endl;
		Indent(os);
		os << "{" << endl;
		Tree[1]->OutputOngo(os, "BGosub", 1);
		Indent(os);
		os << "}" << endl;
		break;

	case BAS_S_OPEN:
		//
		// For INPUT/OUTPUT
		//
		if (Tree[1] != NULL)
		{
			Indent(os);
			os << "BasicChannel[";
			Tree[2]->OutputForcedType(VARTYPE_INTEGER, os);
			os << "].";
			if (Tree[1]->Type == BAS_S_INPUT)
			{
				os << "ForInput();" << endl;
			}
			else
			{
				os << "ForOutput();" << endl;
			}

		}

		//
		// Other flags set on open
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->OutputOpenStuff(os, Tree[2]);
		}


		Indent(os);
		os << "BasicChannel[";
		Tree[2]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Open(";
		if (Tree[0] != NULL)
		{
			os << Tree[0];
		}
		os << ");" << endl;
		break;

	case BAS_S_OPTION:
		break;

	case BAS_X_LOR:
		os << Tree[0] << " || " << Tree[1];
		break;

	case BAS_S_OR:
		os << Tree[0] << " | " << Tree[1];
		break;

	case BAS_V_PREDEF:
		os << TextValue;
		break;

	case BAS_S_PRINT:
		OutputPrint(os);
		break;

	case BAS_S_PUT:
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputGetPutOptions(Tree[0], os);
		}
		Indent(os);
		os << "BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Put();" << endl;
		break;

	case BAS_S_RANDOM:
		Indent(os);
		os << "srand(time(0));" << endl;
		break;

	case BAS_S_RCTLO:
		os << "Rctlo()";
		break;

	case BAS_S_RCTLC:
		os << "Rctlc()";
		break;

	case BAS_S_CTRLC:
		os << "Ctrlc()";
		break;

	case BAS_S_NUM:
		os << "Num";
		break;

	case BAS_S_NUM2:
		os << "Num2";
		break;

	case BAS_S_RECOUNT:
		os << "Recount";
		break;

	case BAS_S_STATUS:
		os << "Status";
		break;

	case BAS_S_RECORD:
		Indent(os);
		os << "struct ";
		if (Tree[0] != NULL)
		{
			os << Tree[0];
		}
		os << endl;
		Indent(os);
		os << "{" << endl;

		if (Tree[3] != NULL)
		{
			Level++;
			Tree[3]->OutputRecord(os);
			Level--;
		}

		Indent(os);
		os << "};" << endl;

		break;

	case BAS_S_RESTORE:
	case BAS_S_RESET:
		if (Tree[0] == NULL)
		{
			Indent(os);
			os << "DataList.Reset();" << endl;
		}
		else
		{
			if (Tree[1] != NULL)
			{
				Tree[1]->OutputGetPutOptions(Tree[0], os);
			}
			Indent(os);
			os << "BasicChannel[";
			Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
			os << "].Reset();" << endl;
		}
		break;

	case BAS_S_READ:
		Indent(os);
		os << "DataList.Read(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_RSET:
		assert(Tree[0] != NULL);
		assert(Tree[0]->Type == BAS_N_ASSIGN);
		assert(Tree[0]->Tree[0] != NULL);
		assert(Tree[0]->Tree[1] != NULL);

		Indent(os);
		os << Tree[0]->Tree[0] << ".Rset(" <<
			Tree[0]->Tree[1] << ");" << endl;
		break;

	case BAS_S_RESUME:
		Indent(os);
		if (Tree[0] != NULL)
		{
			os << "OnErrorZero;" << endl;
			Indent(os);
			os << "goto " << Tree[0] << ";" << endl;
		}
		else
		{
			os << "OnResume;" << endl;
		}
		break;

	case BAS_S_RETURN:
		Indent(os);
		os << "BReturn;" << endl << endl;
		break;

	case BAS_S_SELECT:
		assert(Tree[0] != NULL);

		Indent(os);
		os << "switch (" << Tree[0]->NoParen() << ")" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_STOP:
		Indent(os);
		os << "exit(0);" << endl;
		break;

	case BAS_X_STRREF:
		os << Tree[0] << "." << Tree[1];
		break;

	case BAS_N_STRUCTNAME:
		os << TextValue;
		break;

	case BAS_S_SUB:

		//
		// Empty variable list
		//
		assert(Tree[2] != 0);
		Variables->ClearLocal();

		Tree[2]->VariableScanOne(1);
		Variables->Fixup();

		//
		// Output function name
		//
		os << endl;

		Indent(os);
		Tree[2]->OutputDefinition(os, Tree[0], Type);

		//
		// Braces starting function
		//
		os << endl;
		Indent(os);
		os << "{" << endl;

		//
		// Output local variables
		//
		if (Tree[3] != 0)
		{
			Level++;

			Tree[3]->VariableScan(0);
			Variables->Fixup();

			Variables->OutputDef(os, Level);
			os << endl;
			Tree[3]->OutputCode(os);
			Level--;
		}

		Indent(os);
		os << "}" << endl;

		break;

	case BAS_S_SUBEXIT:
		Indent(os);
		os <<  "return;" << endl;
		break;

	case BAS_V_TEXTSTRING:
		os << TextValue;
		break;

	case BAS_P_TITLE:
	case BAS_P_SBTTL:
	case BAS_P_IDENT:
	case BAS_S_REMARK:
		Indent(os);
		OutputRemark(os);
		break;

	case BAS_S_TO:
		//
		// TO used in a case statement
		//
		os << Tree[0] << " ... " << Tree[1];
		break;

	case BAS_N_TYPEFUN:
		os << "(";
		OutputDatatype(os, Tree[1]->Type);
		os << ")(" << Tree[0] << ")";
		break;

	case BAS_N_UMINUS:
		os << "-" << Tree[0];
		break;

	case BAS_N_UPLUS:
		os << "+" << Tree[0];
		break;

	case BAS_X_UNARYGT:
		os << "> " << Tree[0];
		break;

	case BAS_X_UNARYGE:
		os << ">= " << Tree[0];
		break;

	case BAS_X_UNARYLE:
		os << "<= " << Tree[0];
		break;

	case BAS_X_UNARYLT:
		os << "< " << Tree[0];
		break;

	case BAS_X_UNARYNEQ:
		os << "!= " << Tree[0];
		break;

	case BAS_S_UNLESS:
		assert(Tree[0] != NULL);

		Indent(os);
		LookDown = Tree[0];
		while(LookDown->Type == '(')
		{
			assert(Tree[0] != NULL);
			LookDown = LookDown->Tree[0];
		}

		switch(LookDown->Type)
		{
		case '=':
			os << "if (" << LookDown->Tree[0] << " != " <<
				LookDown->Tree[1] << ")" << endl;
			break;

		case '>':
			os << "if (" << LookDown->Tree[0] << " <= " <<
				LookDown->Tree[1] << ")" << endl;
			break;

		case '<':
			os << "if (" << LookDown->Tree[0] << " >= " <<
				LookDown->Tree[1] << ")" << endl;
			break;

		case BAS_X_NEQ:
			os << "if (" << LookDown->Tree[0] << " == " <<
				LookDown->Tree[1] << ")" << endl;
			break;

		case BAS_X_GE:
			os << "if (" << LookDown->Tree[0] << " < " <<
				LookDown->Tree[1] << ")" << endl;
			break;

		case BAS_X_LE:
			os << "if (" << LookDown->Tree[0] << " > " <<
				LookDown->Tree[1] << ")" << endl;
			break;

		case BAS_X_LNOT:
			os << "if (" << LookDown->Tree[0] << ")" << endl;
			break;

		default:
			os << "if (!(" << LookDown << "))" << endl;
			break;
		}

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_UNLOCK:
		Indent(os);
		os << "BasicChannel[";
		Tree[0]->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Unlock();" << endl;
		break;

	case BAS_S_UNTIL:
		assert(Tree[0] != NULL);

		Indent(os);
		os << "while (!(" << Tree[0]->NoParen() << "))" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_UPDATE:
		Indent(os);
		os << "RmsUpdate(" << Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_V_USELABEL:
		OutputLabel(os);
		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		if (Tree[2] != NULL)
		{
			Tree[2]->OutputDefinitionParams(os, this, Type);
		}
		else
		{
			OutputDatatype(os, Type);
		}
		break;

	case BAS_S_WAIT:
	case BAS_S_SLEEP:
		Indent(os);
		os << "sleep(" << Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_S_WHILE:
		assert(Tree[0] != NULL);

		Indent(os);
		os << "while (" << Tree[0]->NoParen() << ")" << endl;

		Tree[3]->OutputBlock(os);
		break;

	case BAS_S_XOR:
		os << Tree[0] << " ^ " << Tree[1];
		break;

	default:
		os << "??" << Type << ":" << TextValue << "(";
		if (Tree[0] != NULL)
		{
			os << Tree[0];
		}
		os << ",";
		if (Tree[1] != NULL)
		{
			os << Tree[1];
		}
		os << ",";
		if (Tree[2] != NULL)
		{
			os << Tree[2];
		}
		os << ",";
		if (Tree[3] != NULL)
		{
			os << Tree[3];
		}
		os << ")" << endl;
		break;
	}
}


/*******************************************************************************

Member Function:

	Node::OutputVarName()

Title:

	Outputs Variable Name

Description:

	Creates the full name for a variable

*******************************************************************************/

void Node::OutputVarName(ostream& os)
{
	//
	// Local Variables
	//
	VariableStruct *ThisVar;

	assert(this != NULL);
	assert(TextValue != NULL);

	//
	// Look for variable in table
	//
	ThisVar = Variables->Lookup(TextValue, Tree[0]);

	//
	// Make up a good name to print for this variable
	//
	if (ThisVar != NULL)
	{
		os << ThisVar->GetName();
	}
	else
	{
		os << TextValue;
	}

	//
	// Handle array references (May need to be modified from x[a,b]
	// to x[a][b] manually.
	//
	if (Tree[0] != NULL)
	{
		if ((ThisVar != NULL) && (ThisVar->Class == VARCLASS_FUNC))
		{
			os << "(" << Tree[0]->NoParen() << ")";
		}
		else
		{
			os << "[";
			Tree[0]->OutputArrayParam(os, 0);
			os << "]";
		}
	}
}

/*******************************************************************************

Member Function:

	Node::OutputPrint()

Title:

	Outputs a print statement.

Description:

	This function will mangle a print statement.

Bugs:


*******************************************************************************/

void Node::OutputPrint(ostream& os)
{
	int ReturnFlag = 0;

	assert(this != NULL);

	//
	// Indent as necessary
	//
	IOChPrinted = 0;	// IO Channel not yet output
	IOChannel = NULL;	// No IO Channel defined yet
	IOUsing = NULL;		// No Print Using Statement yet

	//
	// Output print using
	//
	if (Tree[0] != NULL)
	{
		//
		// Handle print statement
		//
		ReturnFlag = Tree[0]->OutputIPData(os, 0);

		//
		// Finish off print using statement
		//
		if (IOUsing != NULL)
		{
			OutputIPChannel(os, 0);
			os << " << PUse.Finish()";
		}

		//
		// Force a return at end of line?
		//
		if (ReturnFlag != 0)
		{
			OutputIPChannel(os, 0);
			os << " << endl";
		}
	}
	else
	{
		//
		// Empty print command
		//
		OutputIPChannel(os, 0);
		os << " << endl";
	}

	//
	// Close off statement
	//
	if (IOChPrinted != 0)
	{
		os << ";" << endl;
	}
}

/*******************************************************************************
 */
void Node::OutputIPChannel(ostream& os, int InputFlag)
{
	//
	// Output channel only if we haven't already
	//
	if (IOChPrinted == 0)
	{
		Indent(os);

		//
		// Use default if no channel specified, else use
		// whatever is given
		//
		if (IOChannel == NULL)
		{
			if (InputFlag == 0)
			{
				os << "cout";
			}
			else
			{
				os << "cin";
			}
		}
		else
		{
			os << IOChannel;
		}

		//
		// Flag that we have printed the channel
		//
		IOChPrinted = 1;
	}
}

/*******************************************************************************

Member Function:

	Node::OutputInput()

Title:

	Outputs a input statement.

Description:

	This function will mangle a input statement.

Bugs:


*******************************************************************************/

void Node::OutputInput(ostream& os, int InputFlag)
{
	Node *PChannel = 0;
	Node *TreeDown;

	assert(this != NULL);

	//
	// Initialize flags
	//
	IOChPrinted = 0;		// Channel not yet printed
	IOChannel = NULL;		// No channel specified
	IOUsing = NULL;			// No print using specified

	//
	// Write out input statements
	//
	if (Tree[0] != NULL)
	{
		Tree[0]->OutputIPData(os, InputFlag);
	}

	//
	// Finish off input statement
	//
	if (IOChPrinted != 0)
	{
		os << ";" << endl;
	}
}

/*******************************************************************************

Member Function:

	int Node::OutputIPData()

Title:

	Outputs data related to an input or print statement.

Description:

	This function will mangle a input statement. Returns a
	flag depending on if the input ended in something that should
	cause a return to be appended.

Parameters:

	InputFlag

		Specifies if it is working on an input statement,
		ie. specifies the direction of the arrows.

Return:

	0 - Blank item on right hand side.
	1 - Normal item on right hand side.
	2 - Special item was just output

Bugs:


*******************************************************************************/

int Node::OutputIPData(ostream& os, int InputFlag)
{
	int ReturnFlag;		// Does a return seem to be needed
				//   Gets set to one if the print/Input command
				//   ends with a ',' or ';'.
	int TestFlag = 0;	// Test for special item

	assert(this != NULL);

	//
	// Is this a piece of a list?
	//
	switch (Type)
	{
	case BAS_N_LIST:
		//
		// Try the left side of the tree
		//
		if (Tree[0] != NULL)
		{
			TestFlag = Tree[0]->OutputIPData(os, InputFlag);
		}

		//
		// Do we need to append a tab between fields?
		// Only do so for print statements.
		//
		if ((InputFlag == 0) && (TextValue[0] == ',') &&
			(TestFlag != 2) && (IOUsing == NULL))
		{
			OutputIPChannel(os, InputFlag);
			os << " << '\\t'";
		}

		//
		// Handle right side of the tree
		//
		if (Tree[1] != NULL)
		{
			ReturnFlag = Tree[1]->OutputIPData(os, InputFlag);
		}
		else
		{
			ReturnFlag = 0;
		}

		break;

	case '#':
		IOChannel = this;
		ReturnFlag = 2;
		break;

	case BAS_S_USING:
		IOUsing = Tree[0];
		ReturnFlag = 2;

		Indent(os);
		os << "PUse.SetFormat(" << IOUsing << ");" << endl;

		break;

	case BAS_N_RECORD:
		OutputIPChannel(os, InputFlag);
		if (InputFlag == 0)
		{
			os << " >> ";
		}
		else
		{
			os << " << ";
		}
		os << "RECORD(" << Tree[0] << ")";
		ReturnFlag = 2;
		break;

	case BAS_V_TEXTSTRING:
		//
		// If this is a prompt used for input, close the current
		// input and set up an output prompt.
		//
		if ((InputFlag != 0) && (IOChPrinted != 0))
		{
			os << ";" << endl;
		}

		//
		// If this is an input statement, create the whole
		// statement on one line
		//
		if (InputFlag != 0)
		{
			Indent(os);
			os << "cout << " << this << ";" << endl;
			ReturnFlag = 2;
			IOChPrinted = 0;
		}
		else
		{
			OutputIPChannel(os, InputFlag);

			//
			// Put a comma in print using statements, else
			// out out a '<<'.
			//
			if (IOUsing != NULL)
			{
				os << " << PUse.Output(" << this << ");" << endl;
				IOChPrinted = 0;
			}
			else
			{
				os << " << " << this;
			}
			ReturnFlag = 1;
		}
		break;


	default:

		//
		// Output seperator depending on what state we are in
		//
		switch(InputFlag)
		{
		//
		// Print
		//
		case 0:
			OutputIPChannel(os, InputFlag);

			//
			// Put a comma in print using statements, else
			// out out a '<<'.
			//
			if (IOUsing != NULL)
			{
				os << " << PUse.Output(" << this << ");" << endl;
				IOChPrinted = 0;
			}
			else
			{
				os << " << " << this;
			}
			ReturnFlag = 1;
			break;

		//
		// Regular input
		//
		case 1:
			OutputIPChannel(os, InputFlag);
			os << " >> " << this;
			ReturnFlag = 1;
			break;

		//
		// LINPUT/INPUT LINE
		//
		case 2:
			Indent(os);
			os << "{" << endl;
			Level++;
			Indent(os);
			os << "char TmpInput[MAX_INPUT];" << endl;
			OutputIPChannel(os, InputFlag);
			os << ".getline(TmpInput, MAX_INPUT);" << endl;
			Indent(os);
			os << this << " = TmpInput;" << endl;
			Level--;
			Indent(os);
			os << "}" << endl;
			IOChPrinted = 0;
			ReturnFlag = 2;
			break;
		}

		break;
	}

	//
	// Done
	//
	return ReturnFlag;
}

/*******************************************************************************

Global Function:

	Node::OutputData()

Title:

	Outputs a data statement.

Description:

	This function will mangle a data statement.

Bugs:


*******************************************************************************/

static int DataWidth;

void Node::OutputData(ostream& os)
{
	assert(this != NULL);

	DataWidth = 0;
	Indent(os);
	Tree[0]->OutputDataValue(os);

	//
	// Handle additional data statements attached to this one
	//
	Node* Loop = Tree[1];
	while (Loop != NULL)
	{
		//
		// Let's start on a new line for each new
		// data statement, so we keep more of the
		// original formatting.
		//
		os << "," << endl;
		DataWidth = 0;
		Indent(os);

		Loop->Tree[0]->OutputDataValue(os);
		Loop = Loop->Tree[1];
	}
}


/*******************************************************************************

Local Function:

	Node::OutputDataValue()

Title:

	Outputs one data value in a data statement.

Description:

	This function will mangle one data value for a data statement.
	It forces everything to be a string.

Bugs:


*******************************************************************************/

void Node::OutputDataValue(ostream& os)
{
	assert(this != NULL);

	if (Type == BAS_N_LIST)
	{
		Tree[0]->OutputDataValue(os);
		os << ",";
		DataWidth++;
		if (DataWidth > 60)
		{
			os << endl;
			DataWidth = 0;
			Indent(os);
		}
		Tree[1]->OutputDataValue(os);
	}
	else
	{
		if (Type == BAS_V_TEXTSTRING)
		{
			os << this;
			DataWidth += strlen(this->TextValue);
		}
		else
		{
			//
			// Force quotes ao0und data value
			//
			os << '"' << this  << '"';
			DataWidth += strlen(this->TextValue) + 2;
		}
	}
}

/*******************************************************************************

Global Function:

	Node::OutputMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/

void Node::OutputMap(ostream& os)
{
	assert(this != NULL);
	assert(Tree[0] != NULL);

	VariableStruct* ThisVar = Variables->Lookup(Tree[0]->TextValue, NULL);
	assert(ThisVar != NULL);

	Indent(os);
	os << "extern struct" << endl;

	Indent(os);
	os << "{" << endl;

	Level++;
	if (Tree[1] != NULL)
	{
		Tree[1]->OutputOneMap(os, ThisVar);
	}
	Level--;
	Indent(os);
	os << "} " << ThisVar->GetName() << ";" << endl;
}

/*******************************************************************************

Member Function:

	Node::OutputOneMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/


void Node::OutputOneMap(ostream& os, VariableStruct* MapVar)
{
	assert(this != NULL);

	switch (Type)
	{
	case BAS_N_LIST:

		if (Tree[0] != NULL)
		{
			Tree[0]->OutputOneMap(os, MapVar);
		}
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputOneMap(os, MapVar);
		}
		break;

	case BAS_V_NAME:

		Indent(os);
		os << TextValue << ";" << endl;

		break;

	case BAS_N_VARTYPE:
	case BAS_S_STRING:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_DECIMAL:

		assert(Tree[2] != NULL);

		Indent(os);
		os << TextValue << " " << Tree[2]->TextValue;
		os << ";" << endl;
		break;

	default:
		Indent(os);
		OutputVarName(os);
		os << ";;;" << endl;
		break;
	}

	if (Down != NULL)
	{
		Down->OutputOneMap(os, MapVar);
	}
}

/*******************************************************************************

Global Function:

	Node::ScanMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/

void Node::ScanMap()
{
	assert(this != NULL);
	assert(Tree[0] != NULL);

	VariableStruct* ThisVar = new VariableStruct(Tree[0]->TextValue,
		VARTYPE_STRUCT, VARCLASS_NONE, TRUE);
	Variables->Append(ThisVar);
	Variables->Fixup();

	if (Tree[1] != NULL)
	{
		Tree[1]->ScanOneMap(ThisVar);
	}
}

/*******************************************************************************

Member Function:

	Node::ScanOneMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/


void Node::ScanOneMap(VariableStruct* MapVar)
{
	char NameBuf[128];
	char TempBuf[64];

	VariableStruct* ThisVar;

	assert(this != NULL);

	switch (Type)
	{
	case BAS_N_LIST:

		if (Tree[0] != NULL)
		{
			Tree[0]->ScanOneMap(MapVar);
		}
		if (Tree[1] != NULL)
		{
			Tree[1]->ScanOneMap(MapVar);
		}
		break;

	case BAS_V_NAME:

		ThisVar = new VariableStruct(TextValue,
			0, VARCLASS_NONE, TRUE);
		Variables->Append(ThisVar);

		strcpy(TempBuf, TextValue);
		strlwr(TempBuf);
		sprintf(NameBuf, "%s.%s", MapVar->GetName(), TempBuf);
		ThisVar->SetCName(NameBuf);

		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		assert(Tree[2] != NULL);

		ThisVar = new VariableStruct(Tree[2]->TextValue,
			0, VARCLASS_NONE, TRUE);
		Variables->Append(ThisVar);

		strcpy(TempBuf, Tree[2]->TextValue);
		strlwr(TempBuf);
		sprintf(NameBuf, "%s.%s",
			MapVar->GetName(), TempBuf);
		ThisVar->SetCName(NameBuf);

		break;

	default:
		cerr << "Undefined type "<< Type << endl;
		break;
	}

	if (Down != NULL)
	{
		Down->ScanOneMap(MapVar);
	}
}

/*******************************************************************************

Member Function:

	Node::OutputCaseLabel()

Title:

	Output the "case" selections, one at a time.

Description:


Bugs:


*******************************************************************************/

void Node::OutputCaseLabel(ostream& os)
{
	assert(this != NULL);

	//
	// Handle this if it is a list
	//
	if (Type == BAS_N_LIST)
	{
		Tree[0]->OutputCaseLabel(os);
		Tree[1]->OutputCaseLabel(os);
		return;
	}

	//
	// Output one case statement
	//
	Level--;
	Indent(os);
	os << "case " << this << ":" << endl;
	Level++;
}

/*******************************************************************************

Local Function:

	Indent(os)

Title:

	Indents to current level.

Description:

	This function is used to indent the code.

Bugs:


*******************************************************************************/

void Node::Indent(ostream& os)
{
	int TabLoop;

	for (TabLoop = 1; TabLoop <= Level; TabLoop++)
	{
		os << "\t";
	}
}

/*******************************************************************************

Local Function:

	OutputRemark()

Title:

	Prints out a comment.

Description:

	This function strips off all of the junk from a comment
	and prints it out.

Bugs:


*******************************************************************************/

void Node::OutputRemark(ostream& os)
{
	char* Remark = TextValue;
	int pointer;

	assert(TextValue != NULL);

	//
	// Skip start of remark character
	//
	if (*Remark == '!')
	{
		Remark++;
	}

	//
	// Skip extra white space
	//
	if (*Remark == ' ')
	{
		Remark++;
	}

	//
	// Lose any junk at end of comment
	// (Yes, it does modify the original value)
	//
	while((pointer = strlen(Remark) - 1) >= 0)
	{
		// Line continuation marker
		if (Remark[pointer] == '&')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// New Line
		if (Remark[pointer] == '\n')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Space
		if (Remark[pointer] == ' ')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Tab
		if (Remark[pointer] == '\t')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Exclimation point
		if (Remark[pointer] == '!')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Break out of loop
		break;
	}

	//
	// Dump out remark (and maybe add a space to the front)
	//
	if ((*Remark == '\0') || (*Remark == '\t') || (*Remark == '*'))
	{
		os << "//" << Remark;
	}
	else
	{
		os << "// " << Remark;
	}

	//
	// There may be some extra stuff that needs printing
	//
	if (Tree[0] != NULL)
	{
		os << ' ';
		Tree[0]->OutputCode(os);
	}

	os << endl;

}

/*******************************************************************************

Function:

	OutputDatatype(os, int Type)

Title:

	Prints out a data type (in C++).

Description:

	Converts a BASIC data type into a C++ Data type
Bugs:


*******************************************************************************/

void Node::OutputDatatype(ostream& os, int ThisType)
{
	char* DataType = "";

	switch(ThisType)
	{
	case VARTYPE_VOID:
		DataType = "void";
		break;

	case VARTYPE_LABEL:
		DataType = "label";	// ?? Shouldn't ever see
		break;

	case BAS_S_BYTE:
	case VARTYPE_BYTE:
		DataType = "char";
		break;

	case BAS_S_DOUBLE:
	case VARTYPE_DOUBLE:
		DataType = "Dfloat";
		break;

	case BAS_S_GFLOAT:
	case VARTYPE_GFLOAT:
		DataType = "Gfloat";
		break;

	case BAS_S_HFLOAT:
	case VARTYPE_HFLOAT:
		DataType = "Hfloat";
		break;

	case BAS_S_LONG:
	case VARTYPE_LONG:
		DataType = "long";
		break;

	case BAS_S_WORD:
	case VARTYPE_WORD:
		DataType = "short int";
		break;

	case BAS_S_INTEGER:
	case VARTYPE_INTEGER:
		DataType = "int";
		break;

	case BAS_S_REAL:
	case VARTYPE_REAL:
		switch(RealType)
		{
		case VARTYPE_DOUBLE:
			DataType = "Dfloat";
			break;

		case VARTYPE_GFLOAT:
			DataType = "Gfloat";
			break;

		case VARTYPE_HFLOAT:
			DataType = "Hfloat";
			break;
		}
//		DataType = "float";
		break;

	case BAS_S_STRING:
	case VARTYPE_DYNSTR:
	case VARTYPE_FIXSTR:
		DataType = "BString";
		break;

	case BAS_S_RFA:
	case VARTYPE_RFA:
		DataType = "RmsRfa";
		break;

	case BAS_S_DECIMAL:
		DataType = "Decimal";
		break;

	default:
		DataType = "(Undefined void)";;
	}

	os << DataType << " ";
}

/*******************************************************************************

Function:

	OutputOngo()

Title:

	Prints out a computed goto/gosub statement

Description:

	Converts a BASIC data type into a C++ Data type
Bugs:


*******************************************************************************/

void Node::OutputOngo(ostream& os, const char* OnType, int GoLevel)
{
	switch(Type)
	{
	case BAS_N_LIST:
		Tree[0]->OutputOngo(os, OnType, GoLevel);
		Tree[1]->OutputOngo(os, OnType, GoLevel + 1);
		break;

	case BAS_V_USELABEL:
		Indent(os);
		os << "case " << GoLevel << ":" << endl;
		Level++;
		Indent(os);
		os << OnType;
		if (strcmp(OnType, "goto") == 0)
		{
			os << ' ';
			OutputLabel(os);
			os << ";" << endl;
		}
		else
		{
			os << '(';
			OutputLabel(os);
			os << ");" << endl;
		}
		Indent(os);
		os << "break;" << endl;
		Level--;
		break;

	case BAS_S_OTHERWISE:
		Indent(os);
		os << "default:" << endl;
		Level++;
		Indent(os);
		os << OnType << '(';
		OutputLabel(os);
		os << ");" << endl;
		Indent(os);
		os << "break;" << endl;
		break;

	default:
		os << "Error in computed " << OnType <<
			" (" << this << ")" << endl;
		break;
	}
}

/*******************************************************************************

Function:

	OutputLabel()

Title:

	Prints out a label

Description:

	Output a label (if referenced anywhere)

Bugs:


*******************************************************************************/

void Node::OutputLabel(ostream& os)
{
	VariableStruct *ThisVar = Variables->Lookup(TextValue, 0);
	if (ThisVar != NULL)
	{
		os << ThisVar->GetName();
	}
	else
	{
		os << TextValue;
	}
}

/*******************************************************************************

Function:

	OutputDefinitionList()

Title:

	Prints out a series of definitions

Description:


Bugs:


*******************************************************************************/

void Node::OutputDefinitionList(ostream& os, Node* MainType, int ExtType)
{
	VariableStruct *ThisVar;

	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:

		assert(Tree[0] != NULL);
		assert(Tree[1] != NULL);

		//
		// Put multiple items on seperate lines
		//
		Tree[0]->OutputDefinitionList(os, MainType, ExtType);
		Tree[1]->OutputDefinitionList(os, MainType, ExtType);

		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		//
		// This is a data type, so see if there are any variables
		// listed under this type, else dump out the data type
		// all be itself.
		//
		if (Tree[2] != NULL)
		{
			Tree[2]->OutputDefinitionList(os, this, ExtType);
		}
		else
		{
			Indent(os);
			OutputDefinition(os, this, ExtType);
			os << ";" << endl;
		}
		break;

	default:
		Indent(os);
		OutputDefinition(os, MainType, ExtType);
		os << ";" << endl;
		break;
	}
}

/*******************************************************************************

Function:

	OutputDefinitionParams()

Title:

	Prints out a series of paramater definitions

Description:


Bugs:


*******************************************************************************/

void Node::OutputDefinitionParams(ostream& os, Node* MainType, int ExtType)
{
	VariableStruct *ThisVar;

	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:

		assert(Tree[0] != NULL);
		assert(Tree[1] != NULL);

		//
		// Put multiple items on seperate lines
		//
		Tree[0]->OutputDefinitionParams(os, MainType, ExtType);
		os << ", ";
		Tree[1]->OutputDefinitionParams(os, MainType, ExtType);

		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		//
		// This is a data type, so see if there are any variables
		// listed under this type, else dump out the data type
		// all be itself.
		//
		if (Tree[2] != NULL)
		{
			Tree[2]->OutputDefinitionParams(os, this, ExtType);
		}
		else
		{
			os << "const ";
			OutputDefinition(os, this, ExtType);
		}
		break;

	default:
		os << "const ";
		OutputDefinition(os, MainType, ExtType);
		break;
	}
}

/*******************************************************************************

Function:

	OutputDefinition()

Title:

	Prints out a definition

Description:


Bugs:


*******************************************************************************/

void Node::OutputDefinition(ostream& os, Node* MainType, int ExtType)
{
	VariableStruct *ThisVar;

	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_ASSIGN:

		Tree[0]->OutputDefinition(os, MainType, BAS_N_ASSIGN);
		os << " = " << Tree[1];
		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		//
		// This is a data type, so see if there are any variables
		// listed under this type, else dump out the data type
		// all be itself.
		//
		if (Tree[2] != NULL)
		{
			Tree[2]->OutputDefinition(os, this, ExtType);
		}
		else
		{
			OutputDatatype(os, Type);
		}
		if (Tree[3] != NULL)
		{
			if (Tree[3]->IsSimpleInteger())
			{
				os << "[" << (atoi(Tree[3]->TextValue) + 1)
					<< "]";
			}
			else
			{
				os << "[";
				Tree[3]->OutputForcedType(VARTYPE_INTEGER, os);
				os << " + 1]";
			}
		}
		break;

	default:
		//
		// Handle class
		//
		switch(ExtType)
		{
		case BAS_S_EXTERNAL:
		case BAS_N_EXTERNALFUNCTION:
		case BAS_N_EXTERNALSUB:
			Indent(os);
			os << "extern ";
			break;

		case BAS_N_EXTERNALCONSTANT:
			Indent(os);
			os << "extern const ";
			break;

		case BAS_S_DEF:
			Indent(os);
			os << "static ";
			break;
		}

		switch (Type)
		{
		case BAS_V_NAME:
			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != NULL)
			{
				OutputDatatype(os, ThisVar->Type);
			}
			else
			{
				if (MainType != NULL)
				{
					MainType->OutputDatatype(os,
						MainType->Type);
				}
				else
				{
					os << "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			if (ThisVar != NULL)
			{
				os << ThisVar->GetName();
			}
			else
			{
				os << TextValue;
			}

			//
			// Handle references (Function or array)
			// (May need to be modified from x[a,b]
			// to x[a][b] manually.
			//
			if (Tree[0] != NULL)
			{
				os << "[";
				Tree[0]->OutputArrayParam(os, 1);
				os << "]";
			}
			if (Tree[3] != NULL)
			{
				if (Tree[3]->IsSimpleInteger())
				{
					os << "[" <<
						(atoi(Tree[3]->TextValue) + 1)
						<< "]";
				}
				else
				{
					os << "[";
					Tree[3]->OutputForcedType(VARTYPE_INTEGER, os);
					os << " + 1]";
				}
			}
			break;

		case BAS_V_FUNCTION:
		case BAS_S_SUB:

			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != NULL)
			{
				OutputDatatype(os, ThisVar->Type);
			}
			else
			{
				if (MainType != NULL)
				{
					MainType->OutputDatatype(os,
						MainType->Type);
				}
				else
				{
					os << "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			if (ThisVar != NULL)
			{
				os << ThisVar->GetName();
			}
			else
			{
				os << TextValue;
			}

			//
			// Handle references (Function)
			//
			os << "(";

			if (Tree[0] != NULL)
			{
				Tree[0]->OutputDefinitionParams(os, NULL,
					BAS_N_LIST);
			}
			else
			{
				if ((ExtType == BAS_S_EXTERNAL) ||
					(ExtType == BAS_N_EXTERNALSUB) ||
					(ExtType == BAS_N_EXTERNALFUNCTION))
				{
					os << "...";
				}
				else
				{
					os << "void";
				}
			}

			os << ")";

			break;

		default:
			os << "(??) " << Type << "(??)" << this;
			break;
		}
		break;

	}
}

/*******************************************************************************

Function:

	OutputDefinition()

Title:

	Prints out a definition using the type guven by another definition

Description:

	Writes out a definition for a new variable based upon previous
	definition.

Bugs:


*******************************************************************************/

void Node::OutputDefinitionCopy(ostream& os, Node* MainType,
	int ExtType, const char* NewName)
{
	VariableStruct *ThisVar;

	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_ASSIGN:

		Tree[0]->OutputDefinitionCopy(os, MainType,
			BAS_N_ASSIGN, NewName);
		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		//
		// This is a data type, so see if there are any variables
		// listed under this type, else dump out the data type
		// all be itself.
		//
		if (Tree[2] != NULL)
		{
			Tree[2]->OutputDefinitionCopy(os, this,
				ExtType, NewName);
		}
		else
		{
			OutputDatatype(os, Type);
		}
		break;

	default:
		switch (Type)
		{
		case BAS_V_NAME:
			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != NULL)
			{
				OutputDatatype(os, ThisVar->Type);
			}
			else
			{
				if (MainType != NULL)
				{
					MainType->OutputDatatype(os,
						MainType->Type);
				}
				else
				{
					os << "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			os << NewName;

			break;

		case BAS_V_FUNCTION:
		case BAS_S_SUB:

			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != NULL)
			{
				OutputDatatype(os, ThisVar->Type);
			}
			else
			{
				if (MainType != NULL)
				{
					MainType->OutputDatatype(os,
						MainType->Type);
				}
				else
				{
					os << "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			os << NewName;

			break;

		default:
			os << "(??) " << Type << "(??)" << this;
			break;
		}
		break;

	}
}

/*******************************************************************************

Function:

	OutputRecord()

Title:

	Prints out a record structure

Description:


Bugs:


*******************************************************************************/

void Node::OutputRecord(ostream& os)
{
	assert(this != NULL);

	switch(Type)
	{
	case BAS_S_VARIANT:

		Indent(os);
		os << "union" << endl;
		Indent(os);
		os << "{" << endl;
		Level++;
		Tree[3]->OutputRecord(os);
		Level--;
		Indent(os);
		os << "};" << endl;

		break;

	case BAS_S_CASE:
		//
		// Variant case
		//
		Indent(os);
		os << "struct" << endl;
		Indent(os);
		os << "{" << endl;
		Level++;
		Tree[3]->OutputRecord(os);
		Level--;
		Indent(os);
		os << "};" << endl;

		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_DECIMAL:

		Tree[2]->OutputDefinitionList(os, this, Type);
		break;

	default:
		OutputCode(os);
		break;
	}

	//
	// Any more levels to structure
	//
	if (Down != NULL)
	{
		Down->OutputRecord(os);
	}
}

/*******************************************************************************

Function:

	OutputArrayParam()

Title:

	Prints out an array reference list

Description:


Bugs:


*******************************************************************************/

void Node::OutputArrayParam(ostream& os, int Increment)
{
	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:

		if (Tree[0] != NULL)
		{
			Tree[0]->OutputArrayParam(os, Increment);
		}
		else
		{
			os << "(NULL)";
		}
		os << "][";
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputArrayParam(os, Increment);
		}
		else
		{
			os << "(NULL)";
		}
		break;

	default:

		if (Increment == 1)
		{
			if (IsSimpleInteger())
			{
				os << (atoi(TextValue) + 1);
			}
			else
			{
				OutputForcedType(VARTYPE_INTEGER, os);
				os << " + 1";
			}
		}
		else
		{
			OutputForcedType(VARTYPE_INTEGER, os);
		}
		break;

	}
}


/*******************************************************************************

Member Function:

	OutputOpenStuff()

Title:

	Output code for an 'OPEN' statement

Description:

	This function generates the code necessary to handle
	a full 'RMS' open statement. It generates calls to a class
	for all the elements available in the statement.

*******************************************************************************/

void Node::OutputOpenStuff(ostream& os, Node* Channel)
{
	switch(Type)
	{
	case BAS_S_RECORDSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetRecordSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_CLUSTERSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetClusterSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_FILESIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetFileSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_EXTENDSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetExtendSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_WINDOWSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetWindowSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_BLOCKSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetBlockSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_BUCKETSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetBucketSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_MODE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetMode(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_BUFFER:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetBuffer(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_DEFAULTNAME:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetDefaultName(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_PRIMARY:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetKey(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_ALTERNATE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetKey(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_TEMPORARY:

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetTemporary(True);" << endl;
		break;

	case BAS_S_CONTIGUOUS:

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetContiguous(True);" << endl;
		break;

	case BAS_S_NOREWIND:

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetRewind(False);" << endl;
		break;

	case BAS_S_NOSPAN:

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetSpan(False);" << endl;
		break;

	case BAS_S_SPAN:

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetSpan(True);" << endl;
		break;

	case BAS_S_ORGANIZATION:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetOrginization(" << Tree[0]->TextValue;
		if (Tree[0]->Tree[0] != NULL)
		{
			os << ", " << Tree[0]->Tree[0]->TextValue;
		}
		os << ");" << endl;
		break;

	case BAS_S_ACCESS:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetAccess(" << Tree[0]->TextValue << ");" << endl;
		break;

	case BAS_S_ALLOW:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetAllow(" << Tree[0]->TextValue << ");" << endl;
		break;

	case BAS_S_CONNECT:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetConnect(BasicChannel[" << Tree[0] << "]);" <<
			endl;
		break;

	case BAS_S_MAP:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetMap(" << Tree[0] << ", sizeof(" << Tree[0] <<
			"));" << endl;
		break;

	case BAS_S_RECORDTYPE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetRecordType(" << Tree[0]->TextValue <<
			");" << endl;
		break;

	default:

		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].GoofyOpenStuff(" << this << ");" << endl;
		break;

	}

	//
	// Handle downlink
	//
	if (Down != NULL)
	{
		Down->OutputOpenStuff(os, Channel);
	}
}

/***********************************************************************
 */
void Node::OutputBstringText(ostream& os)
{
	switch(this->Type)
	{
	case BAS_V_TEXTSTRING:
		os << "BString(";
		OutputCodeOne(os);
		os << ")";
		break;

	default:
		OutputCodeOne(os);
		break;
	}
}

/***********************************************************************
 * Member Function:
 *
 *	Node::OutputInteger
 *
 * Title:
 *
 *	Output an integer/number
 *
 * Description:
 *
 *	Used to output integers, while stripping off unecessary stuff
 *	that confuses C++, like leading zeros and trailing percent
 *	signs..
 */
void Node::OutputInteger(ostream& os)
{
	int OFlag = 0;		// Have we output something yet?
	int Ptr = 0;		// What character are we looking at
	char ch;		// Actual character

	//
	// Output all digits in the string
	//
	while ((ch = TextValue[Ptr++]) != '\0')
	{
		if (ch == '0')
		{
			if (OFlag != 0)
			{
				os << ch;
			}
		}
		else
		{
			if (ch != '%')
			{
				if ((ch == '.') && (OFlag == 0))
				{
					os << '0';
				}
				os << ch;
				OFlag = 1;
			}
		}
	}

	//
	// If we didn't output anything, then output a zero
	// because it must have been all zeros.
	//
	if (OFlag == 0)
	{
		os << '0';
	}
}


/***********************************************************************
 * Member Function:
 *
 *	Node::OutputField
 *
 * Title:
 *
 *	Output a field statement
 *
 * Description:
 *
 *	Used to generate code for field statements
 */
void Node::OutputField(ostream& os)
{
	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:

		assert(Tree[0] != NULL);
		assert(Tree[1] != NULL);

		Tree[0]->OutputField(os);
		Tree[1]->OutputField(os);

		break;

	case BAS_S_AS:
		Indent(os);
		os << Tree[1] << ".ReMap(FieldBase, " << Tree[0] <<
			");" << endl;
		Indent(os);
		os << "FieldBase += " << Tree[0] << ";" << endl;
		break;

	default:
		os << "// **** This shouldn't be possible *****" << endl;
		Indent(os);
		os << this;
		break;
	}
}

/***********************************************************************
 * Member Function:
 *
 *	Node::IsReallyString
 *
 * Title:
 *
 *	Determines if a text string is really a simple string.
 *
 * Description:
 *
 *	Sometines a TEXTSTRING is actually a number (ie "15"C), and
 *	this function is used when a determination is necessary
 *	about the "really a string" quality is needed.
 */
int Node::IsReallyString(void)
{
	if (Type == BAS_V_TEXTSTRING)
	{
		assert(TextValue != NULL);

		int Length = strlen(TextValue);
		if (TextValue[Length - 1] == '"')
		{
			return 1;
		}
	}
	return 0;
}

/***********************************************************************
 * IsLogical
 *
 *	Is the node a Logical operation (Returns true/false indication)
 *
 * Description:
 *
 *	This function is used to determine if a nodes is a logical
 *	operation (returning true/faslse), so that higher code can
 *	optimize itself to also be logical.
 */
int Node::IsLogical()
{
	switch(Type)
	{
	case '<':
	case '>':
	case BAS_X_NEQ:
	case BAS_X_GE:
	case BAS_X_LE:
	case '=':
	case BAS_X_LNOT:
	case BAS_X_LAND:
	case BAS_X_LOR:
		//
		// logical operations
		//
		return 1;

	case '(':
		//
		// Look beneath the skin
		//
		return Tree[0]->IsLogical();

	default:
		//
		// By default, things are not logical
		//
		return 0;
	}
}

/***********************************************************************
 * IsSimpleInteger
 *
 *	Is the node a simple integer value
 *
 * Description:
 *
 *	This function is used to determine if the node is a simple
 *	integer value (i.e. one which we can do a 'atoi' on and get
 *	correct results).
 *
 *	Used mostly for the array size definition routines when trying
 *	to increase the dimension sizes by one.
 */
int Node::IsSimpleInteger()
{
	//
	// Base type must be integer
	//
	if ((Type == BAS_V_INTEGER) || (Type == BAS_V_INT))
	{
		//
		// Has to start with a digit
		//
		if (isdigit(*TextValue))
		{
			//
			// has to end with a digit or a '%'
			//
			char LastChar = TextValue[strlen(TextValue) - 1];
			if (isdigit(LastChar) || (LastChar == '%'))
			{
				return 1;
			}
		}
	}

	return 0;
}


/***********************************************************************
 * CountParam
 *
 *	Count the number of parameters for a function/array
 *
 * Description:
 *
 *	This function is used to determine the number of parameters
 *	that a function/array uses.
 */
int Node::CountParam()
{
	int Counter;

	switch(Type)
	{
	case BAS_N_LIST:
		if (Tree[0] != NULL)
		{
			Counter = Tree[0]->CountParam();
		}
		else
		{
			Counter = 1;		// Empty parameter
		}

		if (Tree[1] != NULL)
		{
			Counter += Tree[1]->CountParam();
		}
		else
		{
			Counter++;		// Empty parameter
		}
		break;

	default:
		//
		// By default, things are not logical
		//
		Counter = 1;
	}
	return Counter;
}

/***********************************************************************
 * StripLeadZero
 *
 *	Strip leading zero's from line numbers
 *
 * Description:
 *
 *	This function strips leading zero's (0) from line numbers,
 *	since they just make it impossible to match real line numbers
 *	against goto's.
 */
void Node::StripLeadZero()
{
	//
	// Cheesy way to do it, but hopefully there won't be too
	// many leading zero's
	//
	while(*TextValue == '0')
	{
		memmove(TextValue, TextValue+1, strlen(TextValue));
	}
}

/***********************************************************************
 * OutputGetPutOptions
 *
 *	Output code got get/put options
 *
 * Description:
 *
 *	Outputs code for get/put/find for such options as 'record',
 *	'rfa', 'regardless' and other modifiers.
 */
void Node::OutputGetPutOptions(Node* Channel, ostream& os)
{
	switch(Type)
	{
	case BAS_N_LIST:
		if (Tree[0] != NULL)
		{
			Tree[0]->OutputGetPutOptions(Channel, os);
		}
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputGetPutOptions(Channel, os);
		}
		break;

	case BAS_V_RMSRECORD:
	case BAS_S_RECORD:
	case BAS_S_BLOCK:
		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetRecord(" <<
			Tree[0] << ");" << endl;
		break;

	case BAS_S_REGARDLESS:
		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetRegardless();" << endl;
		break;

	case BAS_S_KEY:
		Indent(os);
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetKey(" << Tree[0] << ");" << endl;
		if (Tree[1] != NULL)
		{
			Indent(os);
			os << "BasicChannel[";
			Channel->OutputForcedType(VARTYPE_INTEGER, os);
			os << "].SetKeyMode(";
			switch(Tree[1]->Type)
			{
			case BAS_S_EQ:
				os << "Equal";
				break;
			case BAS_S_GE:
				os << "GreaterEqual";
				break;
			case BAS_S_GT:
				os << "Greater";
				break;

			}
			os << ");" << endl;
			Indent(os);
			os << "BasicChannel[";
			Channel->OutputForcedType(VARTYPE_INTEGER, os);
			os << "].SetKeyValue(" << Tree[2] << ");" << endl;
		}
		break;

	case BAS_S_RFA:
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetRfaValue(" <<
			Tree[0] << ");" << endl;
		break;

	case BAS_S_COUNT:
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].SetCount(" << Tree[0] << ");" << endl;
		break;
	default:
		os << "BasicChannel[";
		Channel->OutputForcedType(VARTYPE_INTEGER, os);
		os << "].Goofy(" << this << ");" << endl;
	}
}

/***********************************************************************
 * GetNodeVarType
 *
 *	Calculate the varible type for a given node.
 *
 * Description:
 *
 *	Determines the variable type for a given node, and returns
 *	the specified type.
 */
int Node::GetNodeVarType()
{
	VariableStruct *ThisVar;

	//
	// Handle this node type
	//
	switch (Type)
	{
	//
	// Takes the type of the rvalue
	//
	case '#':
	case '(':
	case BAS_N_ASSIGN:
	case BAS_N_ASSIGNLIST:
	case BAS_S_BY:
	case BAS_N_FORASSIGN:
	case BAS_N_UMINUS:
	case BAS_N_UPLUS:
		return Tree[0]->GetNodeVarType();

	//
	// Takes the 'higher' of the two types
	//
	case '+':
	case '-':
	case '*':
	case '/':
	case '^':
	case BAS_N_LIST:
		return ChooseVarType(Tree[0]->GetNodeVarType(),
			Tree[1]->GetNodeVarType());

	//
	// Always an integer
	//
	case '=':
	case '>':
	case '<':
	case BAS_X_LAND:
	case BAS_S_AND:
	case BAS_S_EQV:
	case BAS_S_IMP:
	case BAS_S_ERL:
	case BAS_S_ERR:
	case BAS_X_GE:
	case BAS_V_INT:
	case BAS_V_INTEGER:
	case BAS_X_LE:
	case BAS_X_NEQ:
	case BAS_S_NOT:
	case BAS_X_LNOT:
	case BAS_X_LOR:
	case BAS_S_OR:
	case BAS_S_RCTLO:
	case BAS_S_RCTLC:
	case BAS_S_CTRLC:
	case BAS_S_NUM:
	case BAS_S_RECOUNT:
	case BAS_S_STATUS:
	case BAS_X_UNARYGT:
	case BAS_X_UNARYGE:
	case BAS_X_UNARYLE:
	case BAS_X_UNARYLT:
	case BAS_X_UNARYNEQ:
	case BAS_S_XOR:
		return VARTYPE_INTEGER;

	//
	// Always a floating point
	//
	case BAS_V_FLOAT:
	case BAS_S_NUM2:
		return VARTYPE_REAL;

	//
	// Always a string
	//
	case BAS_V_TEXTSTRING:
		return VARTYPE_DYNSTR;

	//
	// Function references
	//
	case BAS_V_FUNCTION:
	case BAS_V_NAME:
		//
		// Look for variable in table
		//
		ThisVar = Variables->Lookup(TextValue, Tree[0]);

		//
		// Print this variable
		//
		if (ThisVar != NULL)
		{
			return ThisVar->Type;
		}
		break;

	//
	// Gonna havta handle these at some time, but for now ...
	//
	case BAS_V_PREDEF:
	case BAS_X_STRREF:
	case BAS_N_STRUCTNAME:
	case BAS_N_TYPEFUN:
	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		break;
	}

	//
	// I don't know!
	//
	return VARTYPE_NONE;
}

/***********************************************************************
 * GetNodeVarType
 *
 *	Calculate the varible type for a given node.
 *
 * Description:
 *
 *	Determines the variable type for a given node, and returns
 *	the specified type.
 */
void Node::OutputForcedType(int VarType, ostream& os)
{
	int ThisType = GetNodeVarType();	// Var type of this node
	int NeedForce = 0;			// Do we need to force the type

	//
	// Determine if we really need to force
	//
	switch(VarType)
	{
	case VARTYPE_INTEGER:
		//
		// Integer type allows for BYTE, INTEGER, LONG
		//
		if ((ThisType != VARTYPE_BYTE) &&
			(ThisType != VARTYPE_WORD) &&
			(ThisType != VARTYPE_INTEGER) &&
			(ThisType != VARTYPE_LONG))
		{
			NeedForce = 1;
		}

	default:
		if (ThisType != VarType)
		{
			NeedForce = 1;
		}
	}

	//
	// Now force out the proper code
	//
	if (NeedForce)
	{
		os << "(" << OutputVarType(VarType) << ")(" <<
			this->NoParen() << ")";
	}
	else
	{
		os << this->NoParen();
	}
}
