/****************************************************************************

Program:

	program.cc
	01.04

Title:

	Transform tree to more usable format

Description:

	This procedure scans a tree created by 'parse.y' and transforms
	items that are not handled well in the initial pass.

	Such items include:
	* 'if ... then ... else ... endif' blocks
	* 'for ... next' blocks
	* 'function ... end-function' blocks
	* 'data' statements merged together
	* etc...


Bugs:


History:

	01.03.01 - 04/03/1997 - Kevin Handy
		Split several functions out of main.cc.

	01.03.02 - 04/06/1997 - Kevin Handy
		Added code to move local functions out
		of main function, and to move data statements
		to the top of a function.

	01.03.03 - 06/04/1997 - Kevin Handy
		Modified to merge DATA statements together.

	01.03.04 - 06/16/1997 - Kevin Handy
		Use BAS_S_MAINFUNCTION for the main function.

	01.04.00 - 06/16/1997 - Kevin Handy
		Set version to 1.4

	01.04.01 - 07/09/1997 - Kevin Handy
		Handle bug with multiple data statements.

*******************************************************************************/

//
// System Include Files
//
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>

//
// Project Include Files
//
#include "basic.h"
#include "nodes.h"
#include "parse.h"

//
// Module #defined Constants
//

//
// Module #defined Functions
//

//
// Global Variables
//

//
// Module Structure Definitions
//

//
// Module Function Prototypes
//
static Node* ScanForLoops(Node* Program);
static Node* MoveFunctions(Node* Program);
static Node* MoveFunctionsOne(Node* Program);

//
// Module Variables
//


/*******************************************************************************

Global Function:

	DoProgram()

Title:

	Main part of translater

Description:

	This function does most of the work after the parser has
	generated the tree.

Bugs:

	Not yet complete.

*******************************************************************************/

void DoProgram(Node *Program)
{

	//
	// Pretend it all started with a FUNCTION INTGER MAIN.
	// This creates a 'main' function for the program.
	//
	Node* BeginProgram = new Node(BAS_S_MAINFUNCTION);
	Node* BeginType = new Node(BAS_S_INTEGER);
	Node* BeginName = new Node(BAS_V_FUNCTION, "MAIN");

	BeginProgram->Link(BeginType, NULL, BeginName);
	BeginProgram->DownLink(Program);

	//
	// Dump out according to flags
	//
	if (DebugDump && (Program != 0))
	{
		BeginProgram->PrintTree();
	}

	//
	// Now, mangle the code to correctly block out
	// the program
	//
	if (PositionDump)
	{
		cerr << "Scanning for loops" << endl;
	}
	BeginProgram = ScanForLoops(BeginProgram);

	if (DebugDumpOne)
	{
		BeginProgram->PrintTree();
	}

	//
	// Ok, now let's move functions/data statements to the
	// top of each module.
	//
	if (PositionDump)
	{
		cerr << "Move Local Functions" << endl;
	}
	BeginProgram = MoveFunctions(BeginProgram);

	//
	// Output the translated code
	//
	if (PositionDump)
	{
		cerr << "Outputing code" << endl;
	}
	BeginProgram->Output(*OutFile);

	//
	// Free up all memory allocated
	//
	delete BeginProgram;
}

/*******************************************************************************

Local Function:

	ScanForLoops()

Title:

	Fix up loops, if-then-else, functions.

Description:

	This function scans through the program, fixing up for-next loops,
	if-then-else statements, while statements, program segments, etc.
	Since the parser doesn't make complete loops for anything except
	modified statements, we need to roll the statements together
	here.

Bugs:

	Goes internally into the tree to restructure the code instead of
	using the interface code.

*******************************************************************************/

static Node *ScanForLoops
(
	Node *Program
)
{
	//
	// Local Variables
	//
	Node *ThisTree = Program;
	Node *NextTree;
	Node *IndentTree[128];
	int IndentType[128];
	int ThisLevel = 0;

	//
	// Start the whole mass going
	//
	assert(Program != NULL);
	IndentTree[0] = NULL;

	//
	// Scan through the program, looking for loops at level 0.
	//
	while (ThisTree != NULL)
	{
		//
		// Remember what the next level will be
		//
		NextTree = ThisTree->GetDown();
		ThisTree->UnDownLink();

		//
		// Preview this node, and decide what to do with it.
		//
		switch(ThisTree->GetType())
		{
		case BAS_S_DATA:
			//
			// Force data statements to main statement level
			// (0 being function headers) whenever possible.
			//
			if (ThisLevel > 0)
			{
				if (IndentTree[1] == 0)
				{
					IndentTree[0]->Link(NULL, NULL, NULL, ThisTree);
					IndentTree[1] = ThisTree;
				}
				else
				{
					IndentTree[1]->DownLink(ThisTree);
				}
				goto reloop;
			}
			break;

		case BAS_S_THEN:
			//
			// We don't need to keep THEN's at all.
			//
			delete ThisTree;
			goto reloop;

		case BAS_S_ELSE:
			//
			// Look for matching IF
			//
			while ((ThisLevel > 0) &&
				((IndentType[ThisLevel-1] != BAS_S_IF) ||
				(IndentTree[ThisLevel-1]->GetTree(2) != NULL)) &&
				((IndentType[ThisLevel-1] != BAS_S_UNLESS) ||
				(IndentTree[ThisLevel-1]->GetTree(2) != NULL)))
			{
				ThisLevel--;
			}

			//
			// We better be in an if statement
			//
			if ((IndentType[ThisLevel-1] == BAS_S_IF) ||
				(IndentType[ThisLevel-1] == BAS_S_UNLESS))
			{
				//
				// Use the ELSE node as a hook to this
				// section.
				//
				assert(IndentTree[ThisLevel-1]->GetTree(2) == NULL);
				IndentTree[ThisLevel-1]->Link(NULL, NULL, ThisTree);
				IndentTree[ThisLevel] = ThisTree;
			}
			else
			{
				cerr << "Missing IF for ELSE" << endl;
			}

			goto reloop;

		case BAS_S_NEXT:
			//
			// Next statement
			//
			// Look for previous "for" or "while". This may cause
			// IF's to terminate early, but if the program is
			// structured correctly we will only go back one
			// level.
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_FOR) &&
				(IndentType[ThisLevel-1] != BAS_S_WHILE) &&
				(IndentType[ThisLevel-1] != BAS_N_FORUNTIL) &&
				(IndentType[ThisLevel-1] != BAS_N_FORWHILE))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep NEXT any longer
			//
			delete ThisTree;
			goto reloop;

		case BAS_N_ENDRECORD:
			//
			// END RECORD statement
			//
			// Look for previous "RECORD".
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_RECORD))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep 'END RECORD' any longer
			//
			delete ThisTree;
			goto reloop;

		case BAS_N_ENDVARIANT:
			//
			// END VARIANT statement
			//
			// Look for previous "VARIANT".
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_VARIANT))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep 'END VARIANT' any longer
			//
			delete ThisTree;
			goto reloop;

		case BAS_S_FUNCTIONEND:
		case BAS_S_SUBEND:
			//
			// End a function/subroutine
			//
			// Force back to lowest level
			//
			ThisLevel = 0;

			//
			// We don't need to keep 'END SUB' any longer
			//
			delete ThisTree;
			goto reloop;

		case BAS_S_FNEND:
			//
			// End an internal function
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_DEF))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep FNEND any longer
			//
			delete ThisTree;
			goto reloop;

		case BAS_V_LABEL:
			//
			// A line number will terminate any open IF
			// statements. I'll assume they are all at the
			// top of the stack.
			//
			if (isdigit(ThisTree->GetText()[0]))
			{
				while ((ThisLevel > 0) &&
					((IndentType[ThisLevel-1] == BAS_S_IF) ||
					(IndentType[ThisLevel-1] == BAS_S_UNLESS)))
				{
					ThisLevel--;
				}
			}

			//
			// We need to keep line number.
			//
			break;

		case BAS_N_ENDIF:
			//
			// end-if
			//
			if ((ThisLevel > 0) &&
				((IndentType[ThisLevel-1] == BAS_S_IF) ||
				(IndentType[ThisLevel-1] == BAS_S_UNLESS)))
			{
				ThisLevel--;
			}

			//
			// Lose the 'END IF' statement.
			//
			delete ThisTree;
			goto reloop;

		case BAS_S_CASE:
		case BAS_N_CASEELSE:
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_SELECT) &&
				(IndentType[ThisLevel-1] != BAS_S_VARIANT))
			{
				ThisLevel--;
			}

			break;

		case BAS_N_ENDSELECT:
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_SELECT))
			{
				ThisLevel--;
			}

			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// Lose the 'END SELECT'
			//
			delete ThisTree;
			goto reloop;

		case BAS_S_MAINFUNCTION:
		case BAS_S_FUNCTION:
		case BAS_S_SUB:
			if (ThisTree->GetTree(3) == NULL)
			{
				//
				// A Function or Subroutine
				//
				ThisLevel = 0;
			}
			break;
		}

		//
		// Connect this node into the current level
		//
		if (IndentTree[ThisLevel] == 0)
		{
			//
			// If we start with a empty level,
			// we need to hook this
			// node down to the previous level.
			//
			if (ThisLevel != 0)
			{
				assert(IndentTree[ThisLevel-1]->GetTree(3) == NULL);
				IndentTree[ThisLevel-1]->Link(NULL, NULL, NULL, ThisTree);
			}
		}
		else
		{
			//
			// Link Nodes to the previous node on same level
			//
			IndentTree[ThisLevel]->DownLink(ThisTree);
		}

		IndentTree[ThisLevel] = ThisTree;

		//
		// Look at this node, and decide what to do with it
		//
		switch(ThisTree->GetType())
		{
		case BAS_S_IF:
		case BAS_S_UNLESS:
		case BAS_S_FUNCTION:
		case BAS_S_MAINFUNCTION:
		case BAS_S_SUB:
		case BAS_S_FOR:
		case BAS_N_FORUNTIL:
		case BAS_N_FORWHILE:
		case BAS_S_UNTIL:
		case BAS_S_WHILE:
		case BAS_S_DEF:

			if (ThisTree->GetTree(3) == NULL)
			{
				IndentType[ThisLevel] = ThisTree->GetType();
				ThisLevel++;
				IndentTree[ThisLevel] = 0;
			}
			break;

		case BAS_S_CASE:
		case BAS_N_CASEELSE:
		case BAS_S_SELECT:
		case BAS_S_RECORD:
		case BAS_S_VARIANT:

			assert(ThisTree->GetTree(3) == NULL);

			IndentType[ThisLevel] = ThisTree->GetType();
			ThisLevel++;
			IndentTree[ThisLevel] = 0;
			break;
		}

		//
		// Point to next node
		//
reloop:
		ThisTree = NextTree;
	}

	//
	// Done. Hand back re-structured tree.
	//
	return(Program);
}

/*******************************************************************************

Local Function:

	Move Functions()

Title:

	Move local functions

Description:

	This function moves local functions and data statements
	to the beginning of the each major function.
	This simplifies the work needed to be done during code
	generation.

Bugs:

	Probibly ought to be part of ScanForLoops, but I tried
	to get it into there a couple of times and failed
	horribly.

*******************************************************************************/

static Node* MoveFunctions(Node* Program)
{
	Node* FinalTree = NULL;
	Node* ThisTree = Program;

	while (ThisTree != NULL)
	{
		//
		// Keep track of the next node to look at
		//	
		Node* NextTree = ThisTree->GetDown();
		ThisTree->UnDownLink();

		//
		// We'll assume that all the (Level 0) nodes
		// are function definitions, and progrss them
		// accordingly
		//
		ThisTree = MoveFunctionsOne(ThisTree);
		if (FinalTree == NULL)
		{
			FinalTree = ThisTree;
		}
		else
		{
			FinalTree->DownLink(ThisTree);
		}

		//
		// Go to next node
		//
		ThisTree = NextTree;
	}
	return FinalTree;
}

//
// Helper function for MoveFunctions()
//
static Node* MoveFunctionsOne(Node* Program)
{
	Node* LocalFunctions = NULL;
	Node* LocalData = NULL;
	Node* LocalCode = NULL;

	Node* ThisCode = Program->GetTree(3);

	while (ThisCode != NULL)
	{
		//
		// Keep track of what the next statement to look at will be
		// and disconnect then statement from the next.
		//
		Node* NextCode = ThisCode->GetDown();
		ThisCode->UnDownLink();

		//
		// Determine which code sequence to attach this
		// node onto.
		//
		switch(ThisCode->GetType())
		{
		case BAS_S_DEF:
			//
			// Attach local functions to the function tree
			//
			if (LocalFunctions == NULL)
			{
				LocalFunctions = ThisCode;
			}
			else
			{
				LocalFunctions->DownLink(ThisCode);
			}
			break;

		case BAS_S_DATA:
			//
			// Attach data statements to the data tree
			//
			if (LocalData == NULL)
			{
				LocalData = ThisCode;
			}
			else
			{
				Node* TailNode = LocalData;
				while (TailNode->GetTree(1) != NULL)
				{
					TailNode = TailNode->GetTree(1);
				}
				TailNode->Link(NULL, ThisCode);
			}
			break;

		default:
			//
			// By default, everything sticks to the main
			// code sequence
			//
			if (LocalCode == NULL)
			{
				LocalCode = ThisCode;
			}
			else
			{
				LocalCode->DownLink(ThisCode);
			}
			break;
		}

		//
		// Go to next node down
		//
		ThisCode = NextCode;
	}

	//
	// Now lets hook the various nodes together, and
	// attach then to the programs code.
	//
	Program->UnLink(3);
	Node* FinalProgram = LocalData->DownLink(
		LocalFunctions->DownLink(
		Program->Link(NULL, NULL, NULL,
		LocalCode)));

	//
	// Return back restructured code
	//
	return FinalProgram;
}

