/*******************************************************************************

Program:

	Variable.h
	01.01

Title:

	Header file for variable table

Description:

	This is the header file for the variable table/

Bugs:

	Lousy. Needs complete re-write

History:

	01.01.02 - 10/22/1996 - Kevin Handy
		Added 'SINGLE', "GFLOAT', HFLOAT' types.

	01.03.01 - 04/19/1997 - Kevin Handy
		Start working on parameters, pass mechanism
		Make use of more C++'isms.

	01.04.01 - 07/06/1997 - Kevin Handy
		Renumber VARTYPE's so they fall in weight order.
		Allows easier determination of conversions.

*******************************************************************************/

//
// Variable.h - Some stuff just in case if isn't defined anywhere else
//
#ifndef _VARIABLE_H_
#define _VARIABLE_H_

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

//
// Include files
//
#include <iostream.h>
#include "nodes.h"

//
// Variable Type Definitions
//
//	These are numbered according to the 'weight' of the
//	variable type. (With NONE being the lowest of the low)
//
static const int VARTYPE_NONE =		0;	// Undefined
static const int VARTYPE_VOID =		1;	// Void (no) return
static const int VARTYPE_STRUCT =	2;	// Structure
static const int VARTYPE_RFA =		3;	// RFA
static const int VARTYPE_LABEL =	4;	// Label
static const int VARTYPE_FIXSTR =	5;	// Fixed length string
static const int VARTYPE_DYNSTR =	6;	// Dynamic String
static const int VARTYPE_BYTE =		7;	// Byte (char)
static const int VARTYPE_WORD =		8;	// 16 bit integer
static const int VARTYPE_INTEGER =	9;	// 16 bit integer
static const int VARTYPE_LONG =		10;	// 32 bit integer
static const int VARTYPE_SINGLE =	11;	// Single precision float
static const int VARTYPE_REAL =		12;	// float type
static const int VARTYPE_DOUBLE =	13;	// double type
static const int VARTYPE_HFLOAT =	14;	// HFloat floating point
static const int VARTYPE_GFLOAT =	15;	// GFloat floating point
static const int VARTYPE_DECIMAL =	16;	// Label

static const int VARCLASS_NONE =	0;	// Regular variable
static const int VARCLASS_ARRAY =	1;	// An Array
static const int VARCLASS_FUNC =	2;	// Function
static const int VARCLASS_CONST =	3;	// Constant
static const int VARCLASS_LABEL =	4;	// Label

static const int VARMECH_NONE =		0;	// Whatever goes
static const int VARMECH_REF =		1;	// By Reference
static const int VARMECH_DESC =		2;	// By Descriptor
static const int VARMECH_VALUE =	3;	// By Value
static const int VARMECH_XREF =		4;	// Use C++ '&' reference mechanism (??)

//
// Class for handling basic variable classifications
//
class VariableType
{
public:
	int Type;			// Type (integer, real, ...)
	int Class;			// Variable Class (function, array, ...
	int Mech;			// Passing Mechanism
	int ParCount;			// Number of parameters/array

public:
	VariableType() { Type = VARTYPE_NONE; Class = VARCLASS_NONE;
		Mech = VARMECH_NONE; ParCount = 0; }
	VariableType(int NewType, int NewClass = VARCLASS_NONE,
		int NewMech = VARMECH_NONE)
		{ Type = NewType; Class = NewClass; Mech = NewMech;
		ParCount = 0;}

	void SetType(int NewType) { Type = NewType; }
	void SetClass(int NewClass) { Class = NewClass; }
	void SetMech(int NewMech) { Mech = NewMech; }
	void SetParCount(int NewPar) { ParCount = NewPar; }
};

//
// Structure defining variables
//
class VariableStruct
{
public:
	char *BasicName;		// Name used in BASIC
	int Type;			// Type (integer, real, ...)
	int Class;			// Variable Class (function, array, ...)
//	VariableType VarType;		// What king of variable is this
	char *CName;			// Name to be used in C
	int Output;			// Already output to source?
	struct VariableStruct *Left;	// Pointer to next variable (Less)
	struct VariableStruct *Right;	// Pointer to next variable (Greater)
	int EverUsed;			// Is it ever used (Label)
	int ParCount;			// Parameter count

public:
	VariableStruct(char *VarName = "", int Type = 0,
		int Class = 0, int Output = FALSE, int ParCount = 0);
	~VariableStruct();

	VariableStruct *VariableAdd();
	void Dump();
	void OutputDef(ostream& os, int Level);
	char* GetName() { if (CName != "") return CName;
		else return BasicName; }
	void SetCName(const char* NewName);
	void SetClass(const int NewClass) { Class = NewClass; }

private:
	void Indent(ostream& os, int Level);
};

//
// Class that holds a list of variables
//
class VariableList
{
private:
	VariableStruct* Global;			// Global definitions
	VariableStruct* Local;			// Local definitions

public:
	VariableList();
	~VariableList();

	void VariableList::InitFunctions(void);
	void InitOneFunction(char *BasName, char *CName, int Type);

	VariableList* Append(VariableStruct* Variable, int GlobalFlag = FALSE);
	VariableStruct* Lookup(char *VarName, Node *ArrayRef);
	VariableStruct* LookupCpp(char *VarName);
	int LookupReserved(char *VarName);
	VariableStruct* Define(Node *ThisNode, int GlobalFlag = FALSE,
		int ThisOutput = FALSE);
	void Dump();
	void Fixup();
	void OutputDef(ostream& os, int Level);
	void ClearLocal();

private:
	void Fixup1(VariableStruct* Variable);
	void Fixup2(VariableStruct* Variable);
	void Fixup3(VariableStruct* Variable);
	VariableStruct *LookupCppTree(VariableStruct* Variable, char *VarName);
	void VariableList::DumpTree(VariableStruct* Variable);
	void VariableList::DeleteTree(VariableStruct* Node);
	void VariableList::OutputDefTree(ostream& os,
		VariableStruct* VarTable, int Level);
};

//
// Functions related to variables
//
const char* DumpVarType(int VarType);
const char* OutputVarType(int VarType);
int GuessVarType(const char* Name);

//
// Which type to raise expression to
//
static inline int ChooseVarType(int a, int b) { return (a > b) ? a : b; }

#endif
